"""
Rigid-body dynamics functionality of the Toolbox.

Requires access to:

    * ``links`` list of ``Link`` objects, atttribute
    * ``rne()`` the inverse dynamics method

so must be subclassed by ``DHRobot`` class.

:todo: perhaps these should be abstract properties, methods of this calss
"""
from collections import namedtuple
import numpy as np
from spatialmath.base import \
    getvector, verifymatrix, isscalar, getmatrix, t2r
from scipy import integrate, interpolate
from spatialmath.base import symbolic as sym

from ansitable import ANSITable, Column



class DynamicsMixin:

# --------------------------------------------------------------------- #

    @property
    def gravity(self):
        """
        Get/set default gravitational acceleration (Robot superclass)

        - ``robot.name`` is the default gravitational acceleration

        :return: robot name
        :rtype: ndarray(3,)

        - ``robot.name = ...`` checks and sets default gravitational
          acceleration

        .. note:: If the z-axis is upward, out of the Earth, this should be
            a positive number.
        """
        return self._gravity

    @gravity.setter
    def gravity(self, gravity_new):
        self._gravity = getvector(gravity_new, 3)
        self.dynchanged()

# --------------------------------------------------------------------- #
    def dyntable(self):
        """
        Pretty print the dynamic parameters (Robot superclass)

        The dynamic parameters are printed in a table, with one row per link.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.links[2].dyntable()

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.dyntable()
        """
        table = ANSITable(
            Column("j", colalign=">", headalign="^"),
            Column("m", colalign="<", headalign="^"),
            Column("r", colalign="<", headalign="^"),
            Column("I", colalign="<", headalign="^"),
            Column("Jm", colalign="<", headalign="^"),
            Column("B", colalign="<", headalign="^"),
            Column("Tc", colalign="<", headalign="^"),
            Column("G", colalign="<", headalign="^"), border="thin")

        for j, link in enumerate(self):
            table.row(link.name, *link._dyn2list())
        return str(table)

    def printdyn(self):
        """
        Print dynamic parameters

        Display the kinematic and dynamic parameters to the console in
        reable format
        """
        for j, link in enumerate(self.links):
            print("\nLink {:d}::".format(j), link)
            print(link.dyn(indent=2))

# --------------------------------------------------------------------- #

    def friction(self, qd):
        r"""
        Manipulator joint friction (Robot superclass)

        :param qd: The joint velocities of the robot
        :type qd: ndarray(n)

        :return: The joint friction forces/torques for the robot
        :rtype: ndarray(n,)

        ``robot.friction(qd)`` is a vector of joint friction
        forces/torques for the robot moving with joint velocities ``qd``.

        The friction model includes:

        - Viscous friction which is a linear function of velocity.
        - Coulomb friction which is proportional to sign(qd).

        .. math::

            \tau_j = G^2 B \dot{q}_j + |G_j| \left\{ \begin{array}{ll}
                \tau_{C,j}^+ & \mbox{if $\dot{q}_j > 0$} \\
                \tau_{C,j}^- & \mbox{if $\dot{q}_j < 0$} \end{array} \right.

        .. note::

            - The friction value should be added to the motor output torque to
              determine the nett torque. It has a negative value when qd > 0.
            - The returned friction value is referred to the output of the
              gearbox.
            - The friction parameters in the Link object are referred to the
              motor.
            - Motor viscous friction is scaled up by :math:`G^2`.
            - Motor Coulomb friction is scaled up by math:`G`.
            - The appropriate Coulomb friction value to use in the
              non-symmetric case depends on the sign of the joint velocity,
              not the motor velocity.
            - Coulomb friction is zero for zero joint velocity, stiction is
              not modeled.
            - The absolute value of the gear ratio is used.  Negative gear
              ratios are tricky: the Puma560 robot has negative gear ratio for
              joints 1 and 3.
            - The absolute value of the gear ratio is used. Negative gear
              ratios are tricky: the Puma560 has negative gear ratio for
              joints 1 and 3.

        :seealso: :func:`Robot.nofriction`, :func:`Link.friction`
        """

        qd = getvector(qd, self.n)
        tau = np.zeros(self.n)

        for i in range(self.n):
            tau[i] = self.links[i].friction(qd[i])

        return tau

# --------------------------------------------------------------------- #

    def nofriction(self, coulomb=True, viscous=False):
        """
        Remove manipulator joint friction (Robot superclass)

        :param coulomb: set the Coulomb friction to 0
        :type coulomb: bool
        :param viscous: set the viscous friction to 0
        :type viscous: bool
        :return: A copy of the robot with dynamic parameters perturbed
        :rtype: Robot subclass

        ``nofriction()`` copies the robot and returns
        a robot with the same link parameters except the Coulomb and/or viscous
        friction parameter are set to zero.

        :seealso: :func:`Robot.friction`, :func:`Link.nofriction`
        """

        # shallow copy the robot object
        self.delete_rne()  # remove the inherited C pointers
        nf = self.copy()
        nf.name = 'NF/' + self.name

        # add the modified links (copies)
        nf._links = [link.nofriction(coulomb, viscous) for link in self.links]

        return nf


    def fdyn(
            self, T, q0, torqfun=None, targs=None, qd0=None,
            solver='RK45', sargs=None, dt=None, progress=False):
        """
        Integrate forward dynamics

        :param T: integration time
        :type T: float
        :param q0: initial joint coordinates
        :type q0: array_like
        :param qd0: initial joint velocities, assumed zero if not given
        :type qd0: array_like
        :param torqfun: a function that computes torque as a function of time
        and/or state
        :type torqfun: callable
        :param targs: argumments passed to ``torqfun``
        :type targs: dict
        :type solver: name of scipy solver to use, RK45 is the default
        :param solver: str
        :type sargs: arguments passed to the solver
        :param sargs: dict
        :type dt: time step for results
        :param dt: float
        :param progress: show progress bar, default False
        :type progress: bool

        :return: robot trajectory
        :rtype: namedtuple

        - ``tg = R.fdyn(T, q)`` integrates the dynamics of the robot with zero
          input torques over the time  interval 0 to ``T`` and returns the
          trajectory as a namedtuple with elements:

            - ``t`` the time vector (M,)
            - ``q`` the joint coordinates (M,n)
            - ``qd`` the joint velocities (M,n)

        - ``tg = R.fdyn(T, q, torqfun)`` as above but the torque applied to the
          joints is given by the provided function::

                tau = function(robot, t, q, qd, **args)

        where the inputs are:

            - the robot object
            - current time
            - current joint coordinates (n,)
            - current joint velocity (n,)
            - args, optional keyword arguments can be specified, these are
              passed in from the ``targs`` kewyword argument.

        The function must return a Numpy array (n,) of joint forces/torques.

        Examples:

         #. to apply zero joint torque to the robot without Coulomb
            friction::

                def myfunc(robot, t, q, qd):
                    return np.zeros((robot.n,))

                tg = robot.nofriction().fdyn(5, q0, myfunc)

                plt.figure()
                plt.plot(tg.t, tg.q)
                plt.show()

            We could also use a lambda function::

                tg = robot.nofriction().fdyn(
                    5, q0, lambda r, t, q, qd: np.zeros((r.n,)))

         #. the robot is controlled by a PD controller. We first define a
            function to compute the control which has additional parameters for
            the setpoint and control gains (qstar, P, D)::

                def myfunc(robot, t, q, qd, qstar, P, D):
                    return (qstar - q) * P + qd * D  # P, D are (6,)

                targs = {'qstar': VALUE, 'P': VALUE, 'D': VALUE}
                tg = robot.fdyn(10, q0, myfunc, targs=targs) )

        Many integrators have variable step length which is problematic if we
        want to animate the result.  If ``dt`` is specified then the solver
        results are interpolated in time steps of ``dt``.

        .. note::

            - This function performs poorly with non-linear joint friction,
            such as Coulomb friction.  The R.nofriction() method can be used
            to set this friction to zero.
            - If the function is not specified then zero force/torque is
            applied to the manipulator joints.
            - Interpolation is performed using `ScipY integrate.ode
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html>`
            - The SciPy RK45 integrator is used by default
            - Interpolation is performed using `SciPy interp1
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html>`

        :seealso: :func:`DHRobot.accel`, :func:`DHRobot.nofriction`,
            :func:`DHRobot.rne`.
        """

        n = self.n

        if not isscalar(T):
            raise ValueError('T must be a scalar')
        q0 = getvector(q0, n)
        if qd0 is None:
            qd0 = np.zeros((n,))
        else:
            qd0 = getvector(qd0, n)
        if torqfun is not None:
            if not callable(torqfun):
                raise ValueError('torque function must be callable')
        if sargs is None:
            sargs = {}
        if targs is None:
            targs = {}

        # concatenate q and qd into the initial state vector
        x0 = np.r_[q0, qd0]

        # get user specified integrator
        scipy_integrator = integrate.__dict__[solver]

        integrator = scipy_integrator(
            lambda t, y: self._fdyn(t, y, torqfun, targs),
            t0=0.0, y0=x0, t_bound=T, **sargs
            )

        # initialize list of time and states
        tlist = [0]
        xlist = [np.r_[q0, qd0]]

        if progress:
            _printProgressBar(
                0, prefix='Progress:', suffix='complete', length=60)

        while integrator.status == 'running':

            # step the integrator, calls _fdyn multiple times
            integrator.step()

            if integrator.status == 'failed':
                raise RuntimeError('integration completed with failed status ')

            # stash the results
            tlist.append(integrator.t)
            xlist.append(integrator.y)

            # update the progress bar
            if progress:
                _printProgressBar(
                    integrator.t / T, prefix='Progress:', suffix='complete',
                    length=60)

        # cleanup the progress bar
        if progress:
            print('\r' + ' ' * 90 + '\r')

        tarray = np.array(tlist)
        xarray = np.array(xlist)

        if dt is not None:
            # interpolate data to equal time steps of dt
            interp = interpolate.interp1d(tarray, xarray, axis=0)

            tnew = np.arange(0, T, dt)
            xnew = interp(tnew)
            return namedtuple('fdyn', 't q qd')(tnew, xnew[:, :n], xnew[:, n:])
        else:
            return namedtuple('fdyn', 't q qd')(
                tarray, xarray[:, :n], xarray[:, n:])

    def _fdyn(self, t, x, torqfun, targs):
        """
        Private function called by fdyn

        :param t: current time
        :type t: float
        :param x: current state [q, qd]
        :type x: numpy array (2n,)
        :param torqfun: a function that computes torque as a function of time
        and/or state
        :type torqfun: callable
        :param targs: argumments passed to ``torqfun``
        :type targs: dict

        :return: derivative of current state [qd, qdd]
        :rtype: numpy array (2n,)

        Called by ``fdyn`` to evaluate the robot velocity and acceleration for
        forward dynamics.
        """
        n = self.n

        q = x[0:n]
        qd = x[n:]

        # evaluate the torque function if one is given
        if torqfun is None:
            tau = np.zeros((n,))
        else:
            tau = torqfun(self, t, q, qd, **targs)
            if len(tau) != n or not all(np.isreal(tau)):
                raise RuntimeError(
                    'torque function must return vector with N real elements')

        qdd = self.accel(q, qd, tau)

        return np.r_[qd, qdd]

    def accel(self, q, qd, torque):
        r"""
        Compute acceleration due to applied torque

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param qd: Joint velocity
        :type qd: ndarray(n)
        :param torque: Joint torques of the robot
        :type torque:  ndarray(n)
        :return: Joint accelerations of the robot
        :rtype: ndarray(n)

        ``qdd = accel(q, qd, torque)`` calculates a vector (n) of joint
        accelerations that result from applying the actuator force/torque (n)
        to the manipulator in state `q` (n) and `qd` (n), and ``n`` is
        the number of robot joints.

        :math:`\ddot{q} = \mathbf{I}^{-1} \left(\tau - \mathbf{C}(q)\dot{q} - \mathbf{g}(q)\right)`

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.accel(puma.qz, 0.5 * np.ones(6), np.zeros(6))

        **Trajectory operation**

        If `q`, `qd`, torque are matrices (m,n) then ``qdd`` is a matrix (m,n)
        where each row is the acceleration corresponding to the equivalent cols
        of q, qd, torque.

        .. note::
            - Useful for simulation of manipulator dynamics, in
              conjunction with a numerical integration function.
            - Uses the method 1 of Walker and Orin to compute the forward
              dynamics.
            - Featherstone's method is more efficient for robots with large
              numbers of joints.
            - Joint friction is considered.

        :references:
            - Efficient dynamic computer simulation of robotic mechanisms,
              M. W. Walker and D. E. Orin,
              ASME Journa of Dynamic Systems, Measurement and Control, vol.
              104, no. 3, pp. 205-211, 1982.

        """  # noqa

        trajn = 1

        try:
            q = getvector(q, self.n, 'row')
            qd = getvector(qd, self.n, 'row')
            torque = getvector(torque, self.n, 'row')
        except ValueError:
            trajn = q.shape[0]
            verifymatrix(q, (trajn, self.n))
            verifymatrix(qd, (trajn, self.n))
            verifymatrix(torque, (trajn, self.n))

        qdd = np.zeros((trajn, self.n))

        for i in range(trajn):
            # Compute current manipulator inertia torques resulting from unit
            # acceleration of each joint with no gravity.
            qI = (np.c_[q[i, :]] @ np.ones((1, self.n))).T
            qdI = np.zeros((self.n, self.n))
            qddI = np.eye(self.n)

            M = self.rne(qI, qdI, qddI, gravity=[0, 0, 0])

            # Compute gravity and coriolis torque torques resulting from zero
            # acceleration at given velocity & with gravity acting.
            tau = self.rne(q[i, :], qd[i, :], np.zeros((1, self.n)))

            inter = np.expand_dims((torque[i, :] - tau), axis=1)

            qdd[i, :] = np.linalg.solve(M, inter).flatten()

        if trajn == 1:
            return qdd[0, :]
        else:
            return qdd

    def pay(self, W, q=None, J=None, frame=1):
        """
        tau = pay(W, J) Returns the generalised joint force/torques due to a
        payload wrench W applied to the end-effector. Where the manipulator
        Jacobian is J (6xn), and n is the number of robot joints.

        tau = pay(W, q, frame) as above but the Jacobian is calculated at pose
        q in the frame given by frame which is 0 for base frame, 1 for
        end-effector frame.

        Uses the formula tau = J'W, where W is a wrench vector applied at the
        end effector, W = [Fx Fy Fz Mx My Mz]'.

        Trajectory operation:
          In the case q is nxm or J is 6xnxm then tau is nxm where each row
          is the generalised force/torque at the pose given by corresponding
          row of q.

        :param W: A wrench vector applied at the end effector,
            W = [Fx Fy Fz Mx My Mz]
        :type W: ndarray(6)
        :param q: Joint coordinates
        :type q: ndarray(n)
        :param J: The manipulator Jacobian (Optional, if not supplied will
            use the q value).
        :type J: ndarray(6,n)
        :param frame: The frame in which to torques are expressed in when J
            is not supplied. 0 means base frame of the robot, 1 means end-
            effector frame
        :type frame: int

        :return: Joint forces/torques due to w
        :rtype: ndarray(n)

        .. note::
            - Wrench vector and Jacobian must be from the same reference
              frame.
            - Tool transforms are taken into consideration when frame=1.
            - Must have a constant wrench - no trajectory support for this
              yet.

        """

        try:
            W = getvector(W, 6)
            trajn = 0
        except ValueError:
            trajn = W.shape[0]
            verifymatrix(W, (trajn, 6))

        if trajn:
            # A trajectory
            if J is not None:
                # Jacobian supplied
                verifymatrix(J, (trajn, 6, self.n))
            else:
                # Use q instead
                verifymatrix(q, (trajn, self.n))
                J = np.zeros((trajn, 6, self.n))
                for i in range(trajn):
                    if frame:
                        J[i, :, :] = self.jacobe(q[i, :])
                    else:
                        J[i, :, :] = self.jacob0(q[i, :])
        else:
            # Single configuration
            if J is not None:
                # Jacobian supplied
                verifymatrix(J, (6, self.n))
            else:
                # Use q instead
                if q is None:
                    q = np.copy(self.q)
                else:
                    q = getvector(q, self.n)

                if frame:
                    J = self.jacobe(q)
                else:
                    J = self.jacob0(q)

        if trajn == 0:
            tau = -J.T @ W
        else:
            tau = np.zeros((trajn, self.n))

            for i in range(trajn):
                tau[i, :] = -J[i, :, :].T @ W[i, :]

        return tau

    def payload(self, m, p=np.zeros(3)):
        """
        payload(m, p) adds payload mass adds a payload with point mass m at
        position p in the end-effector coordinate frame.

        payload(m) adds payload mass adds a payload with point mass m at
        in the end-effector coordinate frame.

        payload(0) removes added payload.

        :param m: mass (kg)
        :type m: float
        :param p: position in end-effector frame
        :type p: ndarray(3,1)

        """

        p = getvector(p, 3, out='col')
        lastlink = self.links[self.n - 1]

        lastlink.m = m
        lastlink.r = p

    def jointdynamics(self, q, qd=None):
        """
        Transfer function of joint actuator

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param qd: Joint velocity
        :type qd: ndarray(n)
        :return: transfer function denominators
        :rtype: list of 2-tuples

        - ``tf = jointdynamics(qd, q)`` calculates a vector of n
          continuous-time transfer functions that represent the transfer
          function 1/(Js+B) for each joint based on the dynamic parameters
          of the robot and the configuration q (n). n is the number of robot
          joints.

        - ``tf = jointdynamics(q, qd)`` as above but include the linearized
          effects of Coulomb friction when operating at joint velocity QD
          (1xN).
        """

        tf = []
        for j, link in enumerate(self.links):

            # compute inertia for this joint
            zero = np.zeros((self.n))
            qdd = np.zeros((self.n))
            qdd[j] = 1
            M = self.rne(q, zero, qdd, gravity=[0, 0, 0])
            J = link.Jm + M[j] / abs(link.G) ** 2

            # compute friction
            B = link.B
            if qd is not None:
                # add linearized Coulomb friction at the operating point
                if qd > 0:
                    B += link.Tc[0] / qd[j]
                elif qd < 0:
                    B += link.Tc[1] / qd[j]
            tf.append((J, B))

        return tf

    def cinertia(self, q=None):
        r"""
        Cartesian manipulator inertia matrix

        :param q: Joint coordinates
        :type q: ndarray(n)

        :return: The inertia matrix
        :rtype: ndarray(6,6)

        ``robot.cinertia(q)`` is the Cartesian (operational space) inertia
        matrix which relates Cartesian force/torque to Cartesian
        acceleration at the joint configuration q.

        :math:`\mathbf{M} = {\mathbf{J}(q)^+}^T \mathbf{I}(q) \mathbf{J}(q)^+

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.cinertia(puma.qz)

        **Trajectory operation**

        If ``q`` is a matrix (m,n), each row is interpretted as a joint state
        vector, and the result is a 3d-matrix (n,n,m) where each plane
        corresponds to the Cartesian inertia for the corresponding
        row of ``q``.

        .. warning:: Assumes that the operational space has 6 DOF.

        :seealso: :func:`inertia`
        """
        q = getmatrix(q, (None, self.n))

        Mt = np.zeros((q.shape[0], 6, 6))

        for k, qk in enumerate(q):
            J = self.jacob0(qk)
            Ji = np.linalg.pinv(J)
            M = self.inertia(qk)
            Mt[k, :, :] = Ji.T @ M @ Ji

        if q.shape[0] == 1:
            return Mt[0, :, :]
        else:
            return Mt

    def inertia(self, q=None):
        """
        DHRobot.INERTIA Manipulator inertia matrix

        :param q: Joint coordinates
        :type q: ndarray(n)

        :return: The inertia matrix
        :rtype: ndarray(n,n)

        ``inertia(q)`` is the symmetric joint inertia matrix (n,n) which
        relates joint torque to joint acceleration for the robot at joint
        configuration q.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.inertia(puma.qz)

        **Trajectory operation**

        If ``q`` is a matrix (m,n), each row is interpretted as a joint state
        vector, and the result is a 3d-matrix (nxnxk) where each plane
        corresponds to the inertia for the corresponding row of q.

        .. note::
            - The diagonal elements ``I[j,j]`` are the inertia seen by joint
              actuator ``j``.
            - The off-diagonal elements ``I[j,k]`` are coupling inertias that
              relate acceleration on joint ``j`` to force/torque on
              joint ``k``.
            - The diagonal terms include the motor inertia reflected through
              the gear ratio.

        :seealso: :func:`cinertia`
        """
        q = getmatrix(q, (None, self.n))

        In = np.zeros((q.shape[0], self.n, self.n))

        for k, qk in enumerate(q):
            In[k, :, :] = self.rne(
                (np.c_[qk] @ np.ones((1, self.n))).T,
                np.zeros((self.n, self.n)),
                np.eye(self.n),
                gravity=[0, 0, 0])

        if q.shape[0] == 1:
            return In[0, :, :]
        else:
            return In

    def coriolis(self, q, qd):
        r"""
        Coriolis and centripetal term

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param qd: Joint velocity
        :type qd: ndarray(n)
        :return: Velocity matrix
        :rtype: ndarray(n,n)

        ``coriolis(q, qd)`` calculates the Coriolis/centripetal matrix (n,n)
        for the robot in configuration ``q`` and velocity ``qd``, where ``n``
        is the number of joints.

        The product :math:`\mathbf{C} \dot{q}` is the vector of joint
        force/torque due to velocity coupling. The diagonal elements are due to
        centripetal effects and the off-diagonal elements are due to Coriolis
        effects. This matrix is also known as the velocity coupling matrix,
        since it describes the disturbance forces on any joint due to
        velocity of all other joints.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.coriolis(puma.qz, 0.5 * np.ones((6,)))

        **Trajectory operation**

        If ``q`` and `qd` are matrices (m,n), each row is interpretted as a
        joint configuration, and the result (n,n,m) is a 3d-matrix where
        each plane corresponds to a row of ``q`` and ``qd``.

        .. note::
            - Joint viscous friction is also a joint force proportional to
              velocity but it is eliminated in the computation of this value.
            - Computationally slow, involves :math:`n^2/2` invocations of RNE.
        """

        q = getmatrix(q, (None, self.n))
        qd = getmatrix(qd, (None, self.n))
        if q.shape[0] != qd.shape[0]:
            raise ValueError('q and qd must have the same number of rows')

        # ensure that friction doesn't enter the mix, it's also a velocity
        # dependent force/torque
        r1 = self.nofriction(True, True)

        C = np.zeros((q.shape[0], self.n, self.n))
        Csq = np.zeros((q.shape[0], self.n, self.n))

        # Find the torques that depend on a single finite joint speed,
        # these are due to the squared (centripetal) terms
        # set QD = [1 0 0 ...] then resulting torque is due to qd_1^2
        for k, qk in enumerate(q):
            for i in range(self.n):
                QD = np.zeros(self.n)
                QD[i] = 1
                tau = r1.rne(
                    qk, QD, np.zeros(self.n), gravity=[0, 0, 0])
                Csq[k, :, i] = Csq[k, :, i] + tau

        # Find the torques that depend on a pair of finite joint speeds,
        # these are due to the product (Coriolis) terms
        # set QD = [1 1 0 ...] then resulting torque is due to
        # qd_1 qd_2 + qd_1^2 + qd_2^2
        for k, (qk, qdk) in enumerate(zip(q, qd)):
            for i in range(self.n):
                for j in range(i + 1, self.n):
                    # Find a product term  qd_i * qd_j
                    QD = np.zeros(self.n)
                    QD[i] = 1
                    QD[j] = 1
                    tau = r1.rne(qk, QD, np.zeros(self.n), gravity=[0, 0, 0])

                    C[k, :, j] = C[k, :, j] + \
                        (tau - Csq[k, :, j] - Csq[k, :, i]) * qdk[i] / 2

                    C[k, :, i] = C[k, :, i] + \
                        (tau - Csq[k, :, j] - Csq[k, :, i]) * qdk[j] / 2

            C[k, :, :] = C[k, :, :] + Csq[k, :, :] @ np.diag(qdk)

        if q.shape[0] == 1:
            return C[0, :, :]
        else:
            return C

    def itorque(self, q, qdd):
        r"""
        Inertia torque

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param qdd: Joint acceleration
        :type qdd: ndarray(n)

        :return: The inertia torque vector
        :rtype: ndarray(n)

        ``itorque(q, qdd)`` is the inertia force/torque vector (n) at
        the specified joint configuration q (n) and acceleration qdd (n), and
        ``n`` is the number of robot joints.
        It is :math:`\mathbf{I}(q) \ddot{q}`.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.itorque(puma.qz, 0.5 * np.ones((6,)))

        **Trajectory operation**

        If ``q`` and ``qdd`` are matrices (m,n), each row is interpretted as a
        joint configuration, and the result is a matrix (m,n) where each row is
        the corresponding joint torques.

        .. note:: If the robot model contains non-zero motor inertia then this
              will be included in the result.

        :seealso: :func:`inertia`
        """

        q = getmatrix(q, (None, self.n))
        qdd = getmatrix(qdd, (None, self.n))
        if q.shape[0] != qdd.shape[0]:
            raise ValueError('q and qdd must have the same number of rows')

        taui = np.zeros((q.shape[0], self.n))

        for k, (qk, qddk) in enumerate(zip(q, qdd)):
            taui[k, :] = self.rne(
                qk, np.zeros(self.n), qddk, gravity=[0, 0, 0])

        if q.shape[0] == 1:
            return taui[0, :]
        else:
            return taui

    def gravload(self, q=None, gravity=None):
        """
        Compute gravity load

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param gravity: Gravitational acceleration (Optional, if not supplied will
            use the stored gravity values).
        :type gravity: ndarray(3)

        :return: The generalised joint force/torques due to gravity
        :rtype: ndarray(n)

        ``robot.gravload(q)`` calculates the joint gravity loading (n) for
        the robot in the joint configuration ``q`` and using the default
        gravitational acceleration specified in the DHRobot object.

        ``robot.gravload(q, gravity=g)`` as above except the gravitational
        acceleration is explicitly specified as ``g``.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.gravload(puma.qz)

        **Trajectory operation**

        If q is a matrix (nxm) each column is interpreted as a joint
        configuration vector, and the result is a matrix (nxm) each column
        being the corresponding joint torques.

        """

        q = getmatrix(q, (None, self.n))

        if gravity is None:
            gravity = self.gravity
        else:
            gravity = getvector(gravity, 3)

        taug = np.zeros((q.shape[0], self.n))
        z = np.zeros(self.n)

        for k, qk in enumerate(q):
            taug[k, :] = self.rne(qk, z, z, gravity=gravity)

        if q.shape[0] == 1:
            return taug[0, :]
        else:
            return taug

    def paycap(self, w, tauR, frame=1, q=None):
        """
        Static payload capacity of a robot

        :param w: The payload wrench
        :type w: ndarray(n)
        :param tauR: Joint torque matrix minimum and maximums
        :type tauR: ndarray(n,2)
        :param frame: The frame in which to torques are expressed in when J
            is not supplied. 'base' means base frame of the robot, 'ee' means
            end-effector frame
        :type frame: str
        :param q: Joint coordinates
        :type q: ndarray(n)

        :return: The maximum permissible payload wrench
        :rtype: ndarray(6)
        :return: Joint index (zero indexed) which hits its
            force/torque limit
        :rtype: int

        ``wmax, joint = paycap(q, w, f, tauR)`` returns the maximum permissible
        payload wrench ``wmax`` (6) applied at the end-effector, and the index
        of the joint (zero indexed) which hits its force/torque limit at that
        wrench. ``q`` (n) is the manipulator pose, ``w`` the payload wrench
        (6), ``f`` the wrench reference frame and tauR (nx2) is a matrix of
        joint forces/torques (first col is maximum, second col minimum).

        **Trajectory operation:**

        In the case q is nxm then wmax is Mx6 and J is Mx1 where the rows are
        the results at the pose given by corresponding row of q.

        .. note::
            - Wrench vector and Jacobian must be from the same reference frame
            - Tool transforms are taken into consideration for frame=1.
        """
        # TODO rewrite
        trajn = 1

        if q is None:
            q = self.q

        try:
            q = getvector(q, self.n, 'row')
            w = getvector(w, 6, 'row')
        except ValueError:
            trajn = q.shape[1]
            verifymatrix(q, (trajn, self.n))
            verifymatrix(w, (trajn, 6))

        verifymatrix(tauR, (self.n, 2))

        wmax = np.zeros((trajn, 6))
        joint = np.zeros(trajn, dtype=np.int)

        for i in range(trajn):
            tauB = self.gravload(q[i, :])

            # tauP = self.rne(
            #     np.zeros(self.n), np.zeros(self.n),
            #     q, grav=[0, 0, 0], fext=w/np.linalg.norm(w))

            tauP = self.pay(
                w[i, :]/np.linalg.norm(w[i, :]), q=q[i, :], frame=frame)

            M = tauP > 0
            m = tauP <= 0

            TAUm = np.ones(self.n)
            TAUM = np.ones(self.n)

            for c in range(self.n):
                TAUM[c] = tauR[c, 0]
                TAUm[c] = tauR[c, 1]

            WM = np.zeros(self.n)
            WM[M] = (TAUM[M] - tauB[M]) / tauP[M]
            WM[m] = (TAUm[m] - tauB[m]) / tauP[m]

            WM[WM == np.NINF] = np.Inf

            wmax[i, :] = WM
            joint[i] = np.argmin(WM)

        if trajn == 1:
            return wmax[0, :], joint[0]
        else:
            return wmax, joint

    def perturb(self, p=0.1):
        """
        Perturb robot parameters

        rp = perturb(p) is a new robot object in which the dynamic parameters
        (link mass and inertia) have been perturbed. The perturbation is
        multiplicative so that values are multiplied by random numbers in the
        interval (1-p) to (1+p). The name string of the perturbed robot is
        prefixed by 'P/'.

        Useful for investigating the robustness of various model-based control
        schemes. For example to vary parameters in the range +/- 10 percent
        is: r2 = puma.perturb(0.1)

        :param p: The percent (+/-) to be perturbed. Default 10%
        :type p: float

        :return: A copy of the robot with dynamic parameters perturbed
        :rtype: DHRobot

        """

        r2 = self.copy()
        r2.name = 'P/' + self.name

        for i in range(self.n):
            s = (2 * np.random.random() - 1) * p + 1
            r2.links[i].m = r2.links[i].m * s

            s = (2 * np.random.random() - 1) * p + 1
            r2.links[i].I = r2.links[i].I * s    # noqa

        return r2



def _printProgressBar(
        fraction, prefix='', suffix='', decimals=1,
        length=50, fill='█', printEnd="\r"):

    percent = ("{0:." + str(decimals) + "f}").format(fraction * 100)
    filledLength = int(length * fraction)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end=printEnd)




if __name__ == "__main__":   # pragma nocover

    import roboticstoolbox as rtb
    # from spatialmath.base import symbolic as sym

    puma = rtb.models.DH.Puma560()

    # for j, link in enumerate(puma):
    #     print(f'joint {j:}::')
    #     print(link.dyn(indent=4))
    #     print()

    # tau = puma.rne_dh(puma.qz, puma.qz, puma.qz)
    # print(tau)
    # tau = puma.rne_dh(np.r_[puma.qz, puma.qz, puma.qz])
    # print(tau)
    # tau = puma.rne_dh([0,0,0,0,0,0],  [0,0,0,0,0,0],  [0,0,0,0,0,0])
    # print(tau)
    # tau = puma.rne_dh([0,0,0,0,0,0,  0,0,0,0,0,0,  0,0,0,0,0,0])
    # print(tau)

    # puma = rtb.models.DH.Puma560(symbolic=True)
    # print(puma)
    # g = sym.symbol('g')
    # puma.gravity = [0, 0, g]
    # q = sym.symbol('q_:6')
    # qd = sym.symbol('qd_:6')
    # qdd = sym.symbol('qdd_:6')

    # tau = puma.rne_dh(q, qd, qdd, debug=False)

    # print(tau[0].coeff(qdd[0]))
    # print(tau[0].expand().coeff(qdd[0]))

    q = puma.qz
    # qd = puma.qz
    # qdd = puma.qz
    ones = np.ones((6,))
    qd = ones
    qdd = ones

    print(puma.rne(q, qd, qdd))
    print(puma.rne_python(q, qd, qdd, debug=False))

    print(puma.gravity)
    print([link.isrevolute() for link in puma])

    # NOT CONVINCED WE NEED THIS, AND IT'S ORPHAN CODE
    # def gravjac(self, q, grav=None):
    #     """
    #     Compute gravity load and Jacobian

    #     :param q: The joint configuration of the robot
    #     :type q: ndarray(n)
    #     :param grav: The gravity vector (Optional, if not supplied will
    #         use the stored gravity values).
    #     :type grav: ndarray(3,)

    #     :return tau: The generalised joint force/torques due to gravity
    #     :rtype tau: ndarray(n,)

    #     ``tauB = gravjac(q, grav)`` calculates the generalised joint force/
    #     torques due to gravity and the Jacobian

    #     Trajectory operation:
    #     If q is nxm where n is the number of robot joints then a
    #     trajectory is assumed where each row of q corresponds to a robot
    #     configuration. tau (nxm) is the generalised joint torque, each row
    #     corresponding to an input pose, and jacob0 (6xnxm) where each
    #     plane is a Jacobian corresponding to an input pose.

    #     .. note::
    #         - The gravity vector is defined by the SerialLink property if not
    #           explicitly given.
    #         - Does not use inverse dynamics function RNE.
    #         - Faster than computing gravity and Jacobian separately.

    #     Written by Bryan Moutrie

    #     :seealso: :func:`gravload`
    #     """

    #     # TODO use np.cross instead
    #     def _cross3(self, a, b):
    #         c = np.zeros(3)
    #         c[2] = a[0] * b[1] - a[1] * b[0]
    #         c[0] = a[1] * b[2] - a[2] * b[1]
    #         c[1] = a[2] * b[0] - a[0] * b[2]
    #         return c

    #     def makeJ(O, A, e, r):
    #         J[3:6,:] = A
    #         for j in range(r):
    #             if r[j]:
    #                 J[0:3,j] = cross3(A(:,j),e-O(:,j));
    #             else:
    #                 J[:,j] = J[[4 5 6 1 2 3],j]; %J(1:3,:) = 0;

    #     if grav is None:
    #         grav = np.copy(self.gravity)
    #     else:
    #         grav = getvector(grav, 3)

    #     try:
    #         if q is not None:
    #             q = getvector(q, self.n, 'col')
    #         else:
    #             q = np.copy(self.q)
    #             q = getvector(q, self.n, 'col')

    #         poses = 1
    #     except ValueError:
    #         poses = q.shape[1]
    #         verifymatrix(q, (self.n, poses))

    #     if not self.mdh:
    #         baseAxis = self.base.a
    #         baseOrigin = self.base.t

    #     tauB = np.zeros((self.n, poses))

    #     # Forces
    #     force = np.zeros((3, self.n))

    #     for joint in range(self.n):
    #         force[:, joint] = np.squeeze(self.links[joint].m * grav)

    #     # Centre of masses (local frames)
    #     r = np.zeros((4, self.n))
    #     for joint in range(self.n):
    #         r[:, joint] = np.r_[np.squeeze(self.links[joint].r), 1]

    #     for pose in range(poses):
    #         com_arr = np.zeros((3, self.n))

    #         T = self.fkine_all(q[:, pose])

    #         jointOrigins = np.zeros((3, self.n))
    #         jointAxes = np.zeros((3, self.n))
    #         for i in range(self.n):
    #             jointOrigins[:, i] = T[i].t
    #             jointAxes[:, i] = T[i].a

    #         if not self.mdh:
    #             jointOrigins = np.c_[
    #                 baseOrigin, jointOrigins[:, :-1]
    #             ]
    #             jointAxes = np.c_[
    #                 baseAxis, jointAxes[:, :-1]
    #             ]

    #         # Backwards recursion
    #         for joint in range(self.n - 1, -1, -1):
    #             # C.o.M. in world frame, homog
    #             com = T[joint].A @ r[:, joint]

    #             # Add it to the distal others
    #             com_arr[:, joint] = com[0:3]

    #             t = np.zeros(3)

    #             # for all links distal to it
    #             for link in range(joint, self.n):
    #                 if not self.links[joint].sigma:
    #                     # Revolute joint
    #                     d = com_arr[:, link] - jointOrigins[:, joint]
    #                     t = t + self._cross3(d, force[:, link])
    #                     # Though r x F would give the applied torque
    #                     # and not the reaction torque, the gravity
    #                     # vector is nominally in the positive z
    #                     # direction, not negative, hence the force is
    #                     # the reaction force
    #                 else:
    #                     # Prismatic joint
    #                     # Force on prismatic joint
    #                     t = t + force[:, link]

    #             tauB[joint, pose] = t.T @ jointAxes[:, joint]

    #     if poses == 1:
    #         return tauB[:, 0]
    #     else:
    #         return tauB



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import numpy as np
from spatialmath.base.argcheck import getvector


class Gripper():

    def __init__(
            self,
            elinks,
            name=''
            ):

        self._n = 0

        self.name = name

        for link in elinks:
            if link.isjoint:
                self._n += 1

        self.q = np.zeros(self.n)
        self._links = elinks

        # assign the joint indices
        if all([link.jindex is None for link in elinks]):

            jindex = [0]  # "mutable integer" hack

            def visit_link(link, jindex):
                # if it's a joint, assign it a jindex and increment it
                if link.isjoint:
                    link.jindex = jindex[0]
                    jindex[0] += 1

            # visit all links in DFS order
            self.dfs_links(
                elinks[0], lambda link: visit_link(link, jindex))

        elif all([link.jindex is not None for link in elinks]):
            # jindex set on all, check they are unique and sequential
            jset = set(range(self.n))
            for link in elinks:
                if link.jindex not in jset:
                    raise ValueError(
                        'gripper joint index {link.jindex} was '
                        'repeated or out of range')
                jset -= set([link.jindex])
            if len(jset) > 0:   # pragma nocover # is impossible
                raise ValueError('gripper joints {jset} were not assigned')
        else:
            # must be a mixture of ELinks with/without jindex
            raise ValueError(
                'all gripper links must have a jindex, or none have a jindex')

    def dfs_links(self, start, func=None):
        """
        Visit all links from start in depth-first order and will apply
        func to each visited link

        :param start: the link to start at
        :type start: ELink
        :param func: An optional function to apply to each link as it is found
        :type func: function

        :returns: A list of links
        :rtype: list of ELink
        """
        visited = []

        def vis_children(link):
            visited.append(link)
            if func is not None:
                func(link)

            for li in link.child:
                if li not in visited:
                    vis_children(li)

        vis_children(start)

        return visited

    @property
    def n(self):
        return self._n

    @property
    def q(self):
        """
        Get/set gripper joint configuration

        - ``gripper.q`` is the gripper joint configuration

        :return: gripper joint configuration
        :rtype: ndarray(n,)

        - ``gripper.q = ...`` checks and sets the joint configuration

        .. note::  ???
        """
        return self._q

    @q.setter
    def q(self, q_new):
        self._q = getvector(q_new, self.n)

# --------------------------------------------------------------------- #

    @property
    def links(self):
        """
        Gripper links

        :return: A list of link objects
        :rtype: list of Link subclass instances

        .. note:: It is probably more concise to index the robot object rather
            than the list of links, ie. the following are equivalent::

                robot.links[i]
                robot[i]
        """
        return self._links

# --------------------------------------------------------------------- #

    @property
    def name(self):
        """
        Gripper name

        :return: The gripper name
        :rtype: string
        """
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = new_name



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import roboticstoolbox as rp
import numpy as np
from roboticstoolbox.backends.Connector import Connector
from roboticstoolbox.backends.PyPlot.RobotPlot2 import RobotPlot2
from roboticstoolbox.backends.PyPlot.EllipsePlot import EllipsePlot

_mpl = False

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Slider
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42
    plt.style.use('ggplot')
    matplotlib.rcParams['font.size'] = 7
    matplotlib.rcParams['lines.linewidth'] = 0.5
    matplotlib.rcParams['xtick.major.size'] = 1.5
    matplotlib.rcParams['ytick.major.size'] = 1.5
    matplotlib.rcParams['axes.labelpad'] = 1
    plt.rc('grid', linestyle="-", color='#dbdbdb')
    _mpl = True
except ImportError:    # pragma nocover
    pass


class PyPlot2(Connector):

    def __init__(self):

        super(PyPlot2, self).__init__()
        self.robots = []
        self.ellipses = []

        if not _mpl:    # pragma nocover
            raise ImportError(
                '\n\nYou do not have matplotlib installed, do:\n'
                'pip install matplotlib\n\n')

    def launch(self, name=None, limits=None):
        '''
        env = launch() launchs a blank 2D matplotlib figure

        '''

        super().launch()

        labels = ['X', 'Y']

        if name is not None:
            self.fig = plt.figure(name)
        else:
            self.fig = plt.figure()

        # Create a 3D axes
        self.ax = self.fig.add_subplot(1, 1, 1)
        self.ax.set_facecolor('white')

        self.ax.set_xbound(-0.5, 0.5)
        self.ax.set_ybound(-0.5, 0.5)

        self.ax.set_xlabel(labels[0])
        self.ax.set_ylabel(labels[1])

        self.ax.autoscale(enable=True, axis='both', tight=False)
        self.ax.axis('equal')

        if limits is not None:
            self.ax.set_xlim([limits[0], limits[1]])
            self.ax.set_ylim([limits[2], limits[3]])

        plt.ion()
        plt.show()

        # Set the signal handler and a 0.1 second plot updater
        # signal.signal(signal.SIGALRM, self._plot_handler)
        # signal.setitimer(signal.ITIMER_REAL, 0.1, 0.1)

    def step(self, dt=50):
        '''
        state = step(args) triggers the external program to make a time step
        of defined time updating the state of the environment as defined by
        the robot's actions.

        The will go through each robot in the list and make them act based on
        their control type (position, velocity, acceleration, or torque). Upon
        acting, the other three of the four control types will be updated in
        the internal state of the robot object. The control type is defined
        by the robot object, and not all robot objects support all control
        types.

        '''

        super().step()

        self._step_robots(dt)

        plt.ioff()
        self._draw_ellipses()
        self._draw_robots()
        plt.ion()

        self._update_robots()

    def reset(self):
        '''
        state = reset() triggers the external program to reset to the
        original state defined by launch

        '''

        super().reset()

    def restart(self):
        '''
        state = restart() triggers the external program to close and relaunch
        to thestate defined by launch

        '''

        super().restart()

    def close(self):
        '''
        close() closes the plot

        '''

        super().close()

        # signal.setitimer(signal.ITIMER_REAL, 0)
        plt.close(self.fig)

    #
    #  Methods to interface with the robots created in other environemnts
    #

    def add(
            self, ob, readonly=False, display=True,
            eeframe=True, name=False):
        '''
        id = add(robot) adds the robot to the external environment. robot must
        be of an appropriate class. This adds a robot object to a list of
        robots which will act upon the step() method being called.

        '''

        super().add()

        if isinstance(ob, rp.DHRobot) or isinstance(ob, rp.ERobot):
            self.robots.append(
                RobotPlot2(
                    ob, self.ax, readonly, display,
                    eeframe, name))
            self.robots[len(self.robots) - 1].draw2()

        elif isinstance(ob, EllipsePlot):
            ob.ax = self.ax
            self.ellipses.append(ob)
            self.ellipses[len(self.ellipses) - 1].draw2()

    def remove(self):
        '''
        id = remove(robot) removes the robot to the external environment.

        '''

        super().remove()

    def hold(self):           # pragma: no cover
        # signal.setitimer(signal.ITIMER_REAL, 0)
        plt.ioff()
        plt.show()

    #
    #  Private methods
    #

    def _step_robots(self, dt):

        for rpl in self.robots:
            robot = rpl.robot

            if rpl.readonly or robot.control_type == 'p':
                pass            # pragma: no cover

            elif robot.control_type == 'v':

                for i in range(robot.n):
                    robot.q[i] += robot.qd[i] * (dt / 1000)

            elif robot.control_type == 'a':  # pragma: no cover
                pass

            else:            # pragma: no cover
                # Should be impossible to reach
                raise ValueError(
                    'Invalid robot.control_type. '
                    'Must be one of \'p\', \'v\', or \'a\'')

    def _update_robots(self):
        pass

    def _draw_robots(self):

        for i in range(len(self.robots)):
            self.robots[i].draw2()

    def _draw_ellipses(self):

        for i in range(len(self.ellipses)):
            self.ellipses[i].draw2()

    # def _plot_handler(self, sig, frame):
    #     plt.pause(0.001)

    def _add_teach_panel(self, robot):
        fig = self.fig

        # Add text to the plots
        def text_trans(text):  # pragma: no cover
            T = robot.fkine()
            t = np.round(T.t, 3)
            r = np.round(T.rpy(), 3)
            text[0].set_text("x: {0}".format(t[0]))
            text[1].set_text("y: {0}".format(t[1]))
            text[2].set_text("yaw: {0}".format(r[2]))

        # Update the self state in mpl and the text
        def update(val, text, robot):  # pragma: no cover
            for i in range(robot.n):
                robot.q[i] = self.sjoint[i].val * np.pi/180

            text_trans(text)

            # Step the environment
            self.step(0)

        fig.subplots_adjust(left=0.38)
        text = []

        x1 = 0.04
        x2 = 0.22
        yh = 0.04
        ym = 0.5 - (robot.n * yh) / 2 + 0.17/2

        self.axjoint = []
        self.sjoint = []

        qlim = np.copy(robot.qlim) * 180/np.pi

        if np.all(qlim == 0):    # pragma nocover
            qlim[0, :] = -180
            qlim[1, :] = 180

        # Set the pose text
        T = robot.fkine()
        t = np.round(T.t, 3)
        r = np.round(T.rpy(), 3)

        fig.text(
            0.02,  1 - ym + 0.25, "End-effector Pose",
            fontsize=9, weight="bold", color="#4f4f4f")
        text.append(fig.text(
            0.03, 1 - ym + 0.20, "x: {0}".format(t[0]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.03, 1 - ym + 0.16, "y: {0}".format(t[1]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.15, 1 - ym + 0.20, "yaw: {0}".format(r[0]),
            fontsize=9, color="#2b2b2b"))
        fig.text(
            0.02,  1 - ym + 0.06, "Joint angles",
            fontsize=9, weight="bold", color="#4f4f4f")

        for i in range(robot.n):
            ymin = (1 - ym) - i * yh
            self.axjoint.append(
                fig.add_axes([x1, ymin, x2, 0.03], facecolor='#dbdbdb'))

            self.sjoint.append(
                Slider(
                    self.axjoint[i], 'q' + str(i),
                    qlim[0, i], qlim[1, i], robot.q[i] * 180/np.pi))

            self.sjoint[i].on_changed(lambda x: update(x, text, robot))



#!/usr/bin/env python

import numpy as np
from spatialmath.base import trotz, transl
from roboticstoolbox import DHRobot, RevoluteDH


class LWR4(DHRobot):
    """
    Class that models a LWR-IV manipulator

    ``LWR4()`` is a class which models a Kuka LWR-IV robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.LWR4()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration
    - qs, arm is stretched out in the X direction
    - qn, arm is at a nominal non-singular configuration

    .. note:: SI units are used.

    :references:

        - http://www.diag.uniroma1.it/~deluca/rob1_en/09_Exercise_DH_KukaLWR4.pdf

    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self):

        # deg = np.pi/180
        mm = 1e-3
        tool_offset = (103) * mm

        flange = 0 * mm
        # d7 = (58.4)*mm

        # This Kuka model is defined using modified
        # Denavit-Hartenberg parameters

        L = [
                RevoluteDH(
                    a=0.0,
                    d=0,
                    alpha=np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                ),

                RevoluteDH(
                    a=0.0,
                    d=0.0,
                    alpha=-np.pi/2,
                    qlim=np.array([-1.7628, 1.7628])
                ),

                RevoluteDH(
                    a=0.0,
                    d=0.40,
                    alpha=-np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                ),

                RevoluteDH(
                    a=0.0,
                    d=0.0,
                    alpha=np.pi/2,
                    qlim=np.array([-3.0718, -0.0698])
                ),

                RevoluteDH(
                    a=0.0,
                    d=0.39,
                    alpha=np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                ),

                RevoluteDH(
                    a=0.0,
                    d=0.0,
                    alpha=-np.pi/2,
                    qlim=np.array([-0.0175, 3.7525])
                ),

                RevoluteDH(
                    a=0.0,
                    d=flange,
                    alpha=0.0,
                    qlim=np.array([-2.8973, 2.8973])
                )
        ]

        tool = transl(0, 0, tool_offset) @  trotz(-np.pi/4)

        super().__init__(
            L,
            name='LWR-IV',
            manufacturer='Kuka',
            tool=tool)

        # tool = xyzrpy_to_trans(0, 0, d7, 0, 0, -np.pi/4)

        self.addconfiguration("qz", [0, 0, 0, 0, 0, 0, 0])


if __name__ == '__main__':   # pragma nocover

    robot = LWR4()
    print(robot)



from pathlib import Path
import sys

def loadmat(filename):
    """
    Load toolbox mat format data file

    :param filename: relative pathname of datafile
    :type filename: str
    :raises ValueError: File does not exist
    :return: contents of mat data file
    :rtype: dict

    Reads a MATLAB format *mat* file which can contain multiple variables, in 
    a binary or ASCII format.  Returns a dict where the keys are the variable
    names and the values are NumPy arrays.

    .. note::
        - Uses SciPy ``io.loadmat`` to do the work.
        - If the filename has no path component, eg. ``map1.mat`, it will be 
          first be looked for in the folder ``roboticstoolbox/data``.
    
    :seealso: :func:`path_to_datafile`
    """
    from scipy.io import loadmat

    return loaddata(filename, loadmat, squeeze_me=True, struct_as_record=False)

def loaddata(filename, handler, **kwargs):
    """
    Load toolbox data file

    :param filename: relative pathname of datafile
    :type filename: str
    :param handler: function to read data
    :type handler: callable
    :raises ValueError: File does not exist
    :return: data object

    Resolves the relative pathname to an absolute name and then invokes the
    data reading function::

        handler(abs_file_name, **kwargs)
    
    .. note:: If the filename has no path component, eg. ``foo.dat``, it will 
        first be looked for in the folder ``roboticstoolbox/data``.

    :seealso: :func:`path_to_datafile`
    """
    path = path_to_datafile(filename)
    return handler(path, **kwargs)

def path_to_datafile(filename):
    """
    Get absolute path to datafile

    :param filename: pathname of datafile
    :type filename: str
    :raises FileNotFoundError: File does not exist
    :return: Absolute path
    :rtype: str

    If ``filename`` contains no path specification eg. ``map1.mat`` it will
    first attempt to locate the file within the ``roboticstoolbox/data``
    folder and if found, return that absolute path.

    Otherwise, ``~`` is expanded, the path made absolute, resolving symlinks
    and the file's existence is tested.

    Example::

        loadmat('map1.mat')        # read ...roboticstoolbox/data/map1.mat
        loadmat('foo.dat')         # read ./foo.dat
        loadmat('~/data/foo.dat')  # read ~/data/foo.dat
    """

    filename = Path(filename)

    if filename.parent == Path():
        # just a filename, no path, assume it is in roboticstoolbox/data

        p = Path(__file__).resolve().parent.parent / 'data' / filename
        if p.exists():
            return str(p.resolve())

    p = filename.expanduser()
    p = p.resolve()
    if not p.exists():
        raise FileNotFoundError(f"File '{p}' does not exist")
    return str(p)

if __name__ == "__main__":

    a = loadmat("map1.mat")
    print(a)
    a = loadmat("roboticstoolbox/data/map1.mat")
    print(a)
    a = loadmat("roboticstoolbox/data/../data/map1.mat")
    print(a)
    a = loadmat("~/code/robotics-toolbox-python/roboticstoolbox/data/map1.mat")
    print(a)



#!/usr/bin/env python

from roboticstoolbox.robot.ERobot import ERobot


class PR2(ERobot):

    def __init__(self):

        args = super().urdf_to_ets_args(
            "pr2_description/robots/pr2.urdf.xacro",
            "pr2_description")

        super(PR2, self).__init__(
            args[0],
            name=args[1])

        self.manufacturer = 'Willow Garage'



#!/usr/bin/env python
"""
@author: Peter Corke
"""

from roboticstoolbox import DHRobot, RevoluteDH, PrismaticDH
from math import pi
import numpy as np


class Stanford(DHRobot):
    """
    Class that models a Stanford arm manipulator

    ``Stanford()`` is a class which models a Unimation Puma560 robot and
    describes its kinematic and dynamic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Stanford()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero joint angle configuration

    .. note::
        - SI units are used.
        - Gear ratios not currently known, though reflected armature inertia
          is known, so gear ratios are set to 1.

    :references:
        - Kinematic data from "Modelling, Trajectory calculation and Servoing
          of a computer controlled arm".  Stanford AIM-177.  Figure 2.3
        - Dynamic data from "Robot manipulators: mathematics, programming and
          control"
          Paul 1981, Tables 6.5, 6.6
        - Dobrotin & Scheinman, "Design of a computer controlled manipulator
          for robot research", IJCAI, 1973.

    .. codeauthor:: Peter Corke

    """

    def __init__(self):

        deg = pi/180
        inch = 0.0254

        L0 = RevoluteDH(
            d=0.412,      # link length (Dennavit-Hartenberg notation)
            a=0,          # link offset (Dennavit-Hartenberg notation)
            alpha=-pi/2,  # link twist (Dennavit-Hartenberg notation)
            # inertia tensor of link with respect to
            # center of mass I = [L_xx, L_yy, L_zz,
            # L_xy, L_yz, L_xz]
            I=[0.276, 0.255, 0.071, 0, 0, 0],
            # distance of ith origin to center of mass [x,y,z]
            # in link reference frame
            r=[0, 0.0175, -0.1105],
            m=9.29,       # mass of link
            Jm=0.953,     # actuator inertia
            G=1,          # gear ratio
            qlim=[-170*deg, 170*deg])    # minimum and maximum joint angle

        L1 = RevoluteDH(
            d=0.154, a=0., alpha=pi/2,
            I=[0.108, 0.018, 0.100, 0, 0, 0],
            r=[0, -1.054,  0],
            m=5.01, Jm=2.193, G=1,
            qlim=[-170*deg, 170*deg])

        L2 = PrismaticDH(
            theta=-pi/2, a=0.0203, alpha=0,
            I=[2.51, 2.51, 0.006, 0, 0, 0],
            r=[0, 0, -6.447],
            m=4.25, Jm=0.782, G=1,
            qlim=[12*inch, (12+38)*inch])

        L3 = RevoluteDH(
            d=0, a=0, alpha=-pi/2,
            I=[0.002, 0.001, 0.001, 0, 0, 0],
            r=[0, 0.092, -0.054],
            m=1.08, Jm=0.106, G=1,
            qlim=[-170*deg, 170*deg])

        L4 = RevoluteDH(
            d=0, a=0, alpha=pi/2,
            I=[0.003, 0.0004, 0, 0, 0, 0],
            r=[0, 0.566, 0.003], m=0.630,
            Jm=0.097, G=1,
            qlim=[-90*deg, 90*deg])

        L5 = RevoluteDH(
            d=0, a=0, alpha=0,
            I=[0.013, 0.013, 0.0003, 0, 0, 0],
            r=[0, 0, 1.554], m=0.51, Jm=0.020,
            G=1,
            qlim=[-170*deg, 170*deg])

        L = [L0, L1, L2, L3, L4, L5]

        super().__init__(
            L,
            name="Stanford arm",
            manufacturer="Victor Scheinman",
            keywords=('dynamics',))

        # zero angles, L shaped pose
        self.addconfiguration("qz", np.array([0, 0, 0, 0, 0, 0]))


if __name__ == '__main__':  # pragma nocover

    stanford = Stanford()
    print(stanford)



import numpy as np
from spatialmath import base

def jacobian_numerical(f, x, dx=1e-8, N=0):
    r"""
    Numerically compute Jacobian of function

    :param f: the function, returns an m-vector
    :type f: callable
    :param x: function argument
    :type x: ndarray(n)
    :param dx: the numerical perturbation, defaults to 1e-8
    :type dx: float, optional
    :param N: function returns SE(N) matrix, defaults to 0
    :type N: int, optional
    :return: Jacobian matrix
    :rtype: ndarray(m,n)

    Computes a numerical approximation to the Jacobian for ``f(x)`` where 
    :math:`f: \mathbb{R}^n \mapsto \mathbb{R}^m`.

    Uses first-order difference :math:`J[:,i] = (f(x + dx) - f(x)) / dx`.

    If ``N`` is 2 or 3, then it is assumed that the function returns
    an SE(N) matrix which is converted into a Jacobian column comprising the
    translational Jacobian followed by the rotational Jacobian.
    """

    Jcol = []
    J0 = f(x)
    I = np.eye(len(x))
    f0 = f(x)
    for i in range(len(x)):
        fi = f(x + I[:,i] * dx)
        Ji = (fi - f0) / dx

        if N > 0:
            t = Ji[:N,N]
            r = base.vex(Ji[:N,:N] @ J0[:N,:N].T)
            Ji = np.r_[t, r]
        
        Jcol.append(Ji)

    return np.c_[Jcol].T


def hessian_numerical(J, x, dx=1e-8):
    r"""
    Numerically compute Hessian of Jacobian function

    :param J: the Jacobian function, returns an ndarray(m,n)
    :type J: callable
    :param x: function argument
    :type x: ndarray(n)
    :param dx: the numerical perturbation, defaults to 1e-8
    :type dx: float, optional
    :return: Hessian matrix
    :rtype: ndarray(m,n,n)

    Computes a numerical approximation to the Hessian for ``J(x)`` where 
    :math:`f: \mathbb{R}^n  \mapsto \mathbb{R}^{m \times n}`

    Uses first-order difference :math:`H[:,:,i] = (J(x + dx) - J(x)) / dx`.
    """

    I = np.eye(len(x))
    Hcol = []
    J0 = J(x)
    for i in range(len(x)):

        Ji = J(x + I[:,i] * dx)
        Hi = (Ji - J0) / dx

        Hcol.append(Hi)
        
    return np.stack(Hcol, axis=2)


if __name__ == "__main__":

    import roboticstoolbox as rtb
    np.set_printoptions(linewidth=120, formatter={'float': lambda x: f"{x:8.4g}" if abs(x) > 1e-10 else f"{0:8.4g}"})

    robot = rtb.models.DH.Puma560()
    q = robot.qn

    J = jacobian_numerical(lambda q: robot.fkine(q).A, q, N=3)
    print(J)
    print(robot.jacob0(q))

    H = hessian_numerical(robot.jacob0, q)
    print(H)
    print(robot.ets().hessian0(q))


#!/usr/bin/env python
"""
@author Jesse Haviland
"""

from roboticstoolbox.backends.Connector import Connector
import roboticstoolbox as rtb
import numpy as np
import spatialmath as sm

import time
import subprocess
import os
# import rospy
# from std_msgs.msg import Float32MultiArray

_ros = None
rospy = None


def _import_ros():     # pragma nocover
    import importlib
    global rospy
    try:
        ros = importlib.import_module('rospy')
    except ImportError:
        print(
            '\nYou must have ROS installed')
        raise


class ROS(Connector):  # pragma nocover
    """

    """
    def __init__(self):
        super(ROS, self).__init__()

        self.robots = []

    #
    #  Basic methods to do with the state of the external program
    #

    def launch(self, roscore=False, ros_master_uri=None, ros_ip=None):
        """
        """

        super().launch()

        # Launch roscore in seperate process
        if roscore:
            print('Launching ROS core\n')
            self.roscore = subprocess.Popen('roscore')

            # Give it some time to launch
            time.sleep(1)

        if ros_master_uri:
            os.environ["ROS_MASTER_URI"] = ros_master_uri

        if ros_ip:
            os.environ["ROS_IP"] = ros_ip

        v = VelPub(rtb.models.Panda())

    def step(self, dt=0.05):
        """
        """

        super().step

    def reset(self):
        """
        """

        super().reset

    def restart(self):
        """
        """

        super().restart

    def close(self):
        """
        """

        super().close()

    def hold(self):
        """
        """

        super().hold()

    #
    #  Methods to interface with the robots created in other environemnts
    #

    def add(
            self, ob):
        """
        """

        super().add()

    def remove(self):
        """
        Remove a robot to the graphical scene

        ``env.remove(robot)`` removes the ``robot`` from the graphical
            environment.
        """

        # TODO - shouldn't this have an id argument? which robot does it remove
        # TODO - it can remove any entity?

        super().remove()


class VelPub:

    def __init__(self, robot):
        rospy.init_node('RTB_VelocityPublisher')
        self.robot = robot
        self.v = np.zeros(robot.n)

        # self.velocity_publisher = rospy.Publisher(
        #     '/joint_velocity',
        #     Float32MultiArray, queue_size=1)

        self.relay()

    # def relay(self):

    #     while True:
    #         data = Float32MultiArray(data=self.robot.q)
    #         self.velocity_publisher.publish(data)



#!/usr/bin/env python
"""
@author: Jesse Haviland
"""

# import numpy as np
from spatialmath import SE3
# from spatialmath.base.argcheck import getvector, verifymatrix, isscalar
import roboticstoolbox as rp
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.Link import Link


class ELink(Link):
    """
    ETS link class

    :param ets: kinematic - The elementary transforms which make up the link
    :type ets: ETS

    :param qlim: joint variable limits [min max]
    :type qlim: float ndarray(2)

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  SE3
    :param I: dynamic - inertia of link with respect to COM
    :type I: float ndarray(3,3)
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction
    :type B: float
    :param Tc: dynamic - motor Coulomb friction (1x2 or 2x1)
    :type Tc: float ndarray(2)
    :param G: dynamic - gear ratio
    :type G: float

    The ELink object holds all information related to a robot link and can form
    a serial-connected chain or a rigid-body tree.

    It inherits from the Link class which provides common functionality such
    as joint and link such as kinematics parameters,
    .

    :references:
        - Kinematic Derivatives using the Elementary Transform Sequence,
          J. Haviland and P. Corke

    :seealso: :class:`Link`, :class:`DHLink`
    """

    def __init__(
            self,
            ets=ETS(),
            v=None,
            parent=None,
            jindex=None,
            **kwargs):

        # process common options
        super(ELink, self).__init__(**kwargs)

        # check we have an ETS
        if isinstance(ets, ETS):
            self._ets = ets
        else:
            raise TypeError(
                'The ets argument must be of type ETS')

        if v is None and len(ets) > 0 and ets[-1].isjoint:
            v = ets.pop()
            if jindex is not None:
                v.jindex = jindex
            elif jindex is None and v.jindex is not None:
                jindex = v.jindex

        # TODO simplify this logic, can be ELink class or None
        if isinstance(parent, list):
            raise TypeError(
                'Only one parent link can be present')
        elif not isinstance(parent, ELink) and parent is not None:
            raise TypeError(
                'Parent must be of type ELink')

        self._parent = parent
        self._child = []
        self._joint_name = None
        self._jindex = jindex

        # Initialise the static transform representing the constant
        # component of the ETS
        self._init_Ts()

        # Check the variable joint
        if v is None:
            self._joint = False
        elif not isinstance(v, ETS):
            raise TypeError('v must be of type ETS')
        elif not v[0].isjoint:
            raise ValueError('v must be a variable ETS')
        elif len(v) > 1:
            raise ValueError(
                "An elementary link can only have one joint variable")
        else:
            self._joint = True

        self._v = v

    def _init_Ts(self):
        # Number of transforms in the ETS excluding the joint variable
        self._M = len(self._ets)

        # Initialise joints
        if isinstance(self._ets, ETS):
            self._Ts = SE3()
            for i in range(self.M):
                if self._ets[i].isjoint:
                    raise ValueError('The transforms in ets must be constant')

                if not isinstance(self._ets[i].T(), SE3):
                    self._Ts *= SE3(self._ets[i].T())
                else:
                    self._Ts *= self._ets[i].T()

        elif isinstance(self._ets, SE3):
            self._Ts = self._ets

    def __repr__(self):
        name = self.__class__.__name__
        if self.name is None:
            s = f"ets={self.ets()}"
        else:
            s = f"{self.name}, ets={self.ets()}"
        if self.parent is not None:
            s += f", parent={self.parent.name}"
        args = [s] + super()._params()
        return name + "(" + ", ".join(args) + ")"

    def __str__(self):
        """
        Pretty prints the ETS Model of the link. Will output angles in degrees

        :return: Pretty print of the robot link
        :rtype: str
        """
        name = self.__class__.__name__

        if self.name is None:
            return f"{name}[{self.ets()}] "
        else:
            if self.parent is None:
                parent = ""
            else:
                parent = f" [{self.parent.name}]"
            return f"{name}[{self.name}({parent}): {self.ets()}] "

    @property
    def v(self):
        """
        Variable part of link ETS

        :return: joint variable transform
        :rtype: ETS instance

        The ETS for each ELink comprises a constant part (possible the
        identity) followed by an optional joint variable transform.
        This property returns the latter.

        .. runblock:: pycon

            >>> from roboticstoolbox import ELink, ETS
            >>> link = ELink( ETS.tz(0.333) * ETS.rx(90, 'deg') * ETS.rz() )
            >>> print(link.v)
        """
        return self._v

    @property
    def Ts(self):
        """
        Constant part of link ETS

        :return: constant part of link transform
        :rtype: SE3 instance

        The ETS for each ELink comprises a constant part (possible the
        identity) followed by an optional joint variable transform.
        This property returns the constant part.  If no constant part
        is given, this returns an identity matrix.

        .. runblock:: pycon

            >>> from roboticstoolbox import ELink, ETS
            >>> link = ELink( ETS.tz(0.333) * ETS.rx(90, 'deg') * ETS.rz() )
            >>> link.Ts
            >>> link = ELink( ETS.rz() )
            >>> link.Ts
        """
        return self._Ts

    @property
    def isjoint(self):
        """
        Test if link has joint

        :return: test if link has a joint
        :rtype: bool

        The ETS for each ELink comprises a constant part (possible the
        identity) followed by an optional joint variable transform.
        This property returns the whether the

        .. runblock:: pycon

            >>> from roboticstoolbox import models
            >>> robot = models.URDF.Panda()
            >>> robot[1].isjoint  # link with joint
            >>> robot[8].isjoint  # static link
        """
        return self._v is not None

    @property
    def jindex(self):
        """
        Get/set joint index

        - ``link.jindex`` is the joint index
            :return: joint index
            :rtype: int
        - ``link.jindex = ...`` checks and sets the joint index

        For a serial-link manipulator the joints are numbered starting at zero
        and increasing sequentially toward the end-effector.  For branched
        mechanisms this is not so straightforward.

        The link's ``jindex`` property specifies the index of its joint
        variable within a vector of joint coordinates.

        .. note:: ``jindex`` values must be a sequence of integers starting
            at zero.
        """
        return self._jindex

    @jindex.setter
    def jindex(self, j):
        self._jindex = j

    # def isrevolute(self):
    #     """
    #     Checks if the joint is of revolute type

    #     :return: Ture if is revolute
    #     :rtype: bool
    #     """
    #     return self.v.isrevolute

    @property
    def isprismatic(self):
        """
        Checks if the joint is of prismatic type

        :return: True if is prismatic
        :rtype: bool
        """
        return self.isjoint and self.v.isprismatic

    @property
    def isrevolute(self):
        """
        Checks if the joint is of revolute type

        :return: True if is revolute
        :rtype: bool
        """
        return self.isjoint and self.v.isrevolute

    # @property
    # def ets(self):
    #     return self._ets

    # @property
    # def parent_name(self):
    #     return self._parent_name

    # @property
    # def child_name(self):
    #     return self._child_name

    @property
    def parent(self):
        """
        Parent link

        :return: Link's parent
        :rtype: ELink instance

        This is a reference to

        .. runblock:: pycon

            >>> from roboticstoolbox import models
            >>> robot = models.URDF.Panda()
            >>> robot[0].parent  # base link has no parent
            >>> robot[1].parent  # second link's parent
        """
        return self._parent

    @property
    def child(self):
        return self._child

    @property
    def M(self):
        return self._M

    @property
    def geometry(self):
        """
        Get/set joint visual geometry

        - ``link.geometry`` is the list of the visual geometries which
            represent the shape of the link
            :return: the visual geometries
            :rtype: list of Shape
        - ``link.geometry = ...`` checks and sets the geometry
        - ``link.geometry.append(...)`` add geometry
        """
        return self._geometry

    @property
    def collision(self):
        """
        Get/set joint collision geometry

        - ``link.collision`` is the list of the collision geometries which
            represent the collidable shape of the link.
            :return: the collision geometries
            :rtype: list of Shape
        - ``link.collision = ...`` checks and sets the collision geometry
        - ``link.collision.append(...)`` add collision geometry

        The collision geometries are what is used to check for collisions.
        """
        return self._collision

    # @r.setter
    # def r(self, T):
    #     if not isinstance(T, SE3):
    #         T = SE3(T)
    #     self._r = T

    def A(self, q=None, fast=False):
        """
        Link transform matrix

        :param q: Joint coordinate (radians or metres). Not required for links
            with no variable
        :type q: float
        :param fast: return NumPy array instead of ``SE3``
        :type param: bool
        :return T: link frame transformation matrix
        :rtype T: SE3 or ndarray(4,4)

        ``LINK.A(q)`` is an SE(3) matrix that describes the rigid-body 
          transformation from the previous to the current link frame to
          the next, which depends on the joint coordinate ``q``.

        """

        if self.isjoint:
            # a variable joint
            if q is None:
                raise ValueError("q is required for variable joints")
            T = self.Ts.A @ self.v.T(q)
        else:
            # a fixed joint
            T = self.Ts.A

        if fast:
            return T
        else:
            return SE3(T, check=False)

    def ets(self):
        if self.v is None:
            return self._ets
        else:
            return self._ets * self.v

    @collision.setter
    def collision(self, coll):
        new_coll = []

        if isinstance(coll, list):
            for gi in coll:
                if isinstance(gi, rp.Shape):
                    new_coll.append(gi)
                else:
                    raise TypeError('Collision must be of Shape class')
        elif isinstance(coll, rp.Shape):
            new_coll.append(coll)
        else:
            raise TypeError('Geometry must be of Shape class or list of Shape')

        self._collision = new_coll

    @geometry.setter
    def geometry(self, geom):
        new_geom = []

        if isinstance(geom, list):
            for gi in geom:
                if isinstance(gi, rp.Shape):
                    new_geom.append(gi)
                else:
                    raise TypeError('Geometry must be of Shape class')
        elif isinstance(geom, rp.Shape):
            new_geom.append(geom)
        else:
            raise TypeError('Geometry must be of Shape class or list of Shape')

        self._geometry = new_geom



"""
Python Vehicle
@Author: Kristian Gibson
@Author: Peter Corke
"""
from abc import ABC, abstractmethod
import warnings
from math import pi, sin, cos, tan, atan2
import numpy as np
from scipy import integrate, linalg, interpolate

import matplotlib.pyplot as plt
from matplotlib import patches
import matplotlib.transforms as mtransforms

from spatialmath import SE2, base
from roboticstoolbox import loaddata


class VehicleAnimation(ABC):
    
    def __init__(self):
        self._object = None
        self._ax = None

    def add(self, ax=None, **kwargs):
        """
        Add vehicle animation to the current plot

        :param ax: Axis to add to, defaults to current axis
        :type ax: Axes, optional

        A reference to the animation object is kept, and it will be deleted
        from the plot when the ``VehicleAnimation`` object is garbage collected.
        """
        if ax is None:
            self._ax = plt.gca()
        else:
            self._ax = ax

        self._add(**kwargs)

    def update(self, x):
        """
        Update the vehicle animation

        :param x: vehicle state
        :type x: ndarray(2) or ndarray(3)

        The graphical depiction of the vehicle position or pose is updated.
        """
        self._update(x)

    def __del__(self):

        if self._object is not None:
            self._object.remove()

# ========================================================================= #

class VehicleMarker(VehicleAnimation):

    def __init__(self, **kwargs):
        """
        Create graphical animation of vehicle as a marker

        :param kwargs: additional arguments passed to matplotlib ``plot``.
        :return: animation object
        :rtype: VehicleAnimation

        Creates an object that can be passed to a ``Vehicle`` subclass to depict
        the moving robot as a simple matplotlib marker during simulation::

            a = VehicleMarker(marker='s', markerfacecolor='b')
            veh = Bicycle(driver=RandomPath(10), animation=a)
            veh.run()

        .. note:: A marker can only indicate vehicle position, not orientation.

        :seealso: :func:`~Vehicle`
        """
        super().__init__()
        if len(kwargs) == 0:
            kwargs = {
                'marker': 'o',
                'markerfacecolor': 'r',
                'markeredgecolor': 'w',
                'markersize': 12
            }
        self._args = kwargs
        

    def _update(self, x):
        self._object.set_xdata(x[0])
        self._object.set_ydata(x[1])

    def _add(self, x=None, **kwargs):
        if x is None:
            x = (0, 0)
        self._object = plt.plot(x[0], x[1], **self._args)[0]

# ========================================================================= #

class VehiclePolygon(VehicleAnimation):

    def __init__(self, shape='car', scale=1, **kwargs):
        """
        Create graphical animation of vehicle as a polygon

        :param shape: polygon shape, defaults to 'car'
        :type shape: str
        :param scale: Length of the vehicle on the plot, defaults to 1
        :type scale: float
        :param kwargs: additional arguments passed to matplotlib patch such as
            ``color`` (face+edge), ``alpha``, ``facecolor``, ``edgecolor``,
            ``linewidth`` etc.
        :raises ValueError: unknown shape name
        :raises TypeError: bad shape argument
        :return: animation object
        :rtype: VehicleAnimation

        Creates an object that can be passed to a ``Vehicle`` subclass to 
        depict the moving robot as a polygon during simulation::

            a = VehiclePolygon('car', color='r')
            veh = Bicycle()
            veh.run(animation=a)

        ``shape`` can be:

            * ``"car"``  a rectangle with a pointy front
            * ``"triangle"`` a triangle
            * an Nx2 NumPy array of vertices, does not have to be closed.

        :seealso: :func:`~Vehicle`
        """
        super().__init__()
        if isinstance(shape, str):

            h = 0.3
            t = 0.8  # start of head taper
            c = 0.5  # centre x coordinate
            w = 1    # width in x direction

            if shape == 'car':
                self._coords = np.array([
                    [-c,     h],
                    [t - c,  h],
                    [w - c,  0],
                    [t - c, -h],
                    [-c,    -h],
                ]).T * scale
            elif shape == 'triangle':
                self._coords = np.array([
                    [-c,  h],
                    [ w,  0],
                    [-c, -h],
                ]).T * scale
            else:
                raise ValueError('unknown vehicle shape name')

        elif isinstance(shape, np.ndarray) and shape.shape[1] == 2:
            self._coords = shape
        else:
            raise TypeError('unknown shape argument')

        self._args = kwargs

    def _add(self, **kwargs):
        # color is fillcolor + edgecolor
        # facecolor if None is default
        self._ax = plt.gca()
        self._object = patches.Polygon(self._coords.T, **self._args)
        self._ax.add_patch(self._object)

    def _update(self, x):
        
        xy = SE2(x) * self._coords
        self._object.set_xy(xy.T)

# ========================================================================= #

class VehicleIcon(VehicleAnimation):

    def __init__(self, filename, origin=None, scale=1, rotation=0):
        """
        Create graphical animation of vehicle as an icon

        :param filename: Standard icon name or a path to an image
        :type filename: str
        :param origin: Origin of the vehicle coordinate frame, defaults to centre
        :type origin: 2-tuple
        :param scale: Length of the vehicle on the plot, defaults to 1
        :type scale: float
        :param rotation: Vehicle icon heading in degrees, defaults to 0
        :type rotation: float
        :raises ValueError: Icon file not found
        :return: animation object
        :rtype: VehicleAnimation

        Creates an object that can be passed to a ``Vehicle`` subclass to 
        depict the moving robot as a polygon during simulation::

            a = VehicleIcon('redcar', scale=2)
            veh = Bicycle()
            veh.run(animation=a)

        ``filename`` can be:

            * ``"greycar"`` a grey and white car (top view)
            * ``"redcar"`` a red car (top view)
            * path to an image file, including extension

        .. image:: ../../roboticstoolbox/data/greycar.png
          :width: 200px
          :align: center

        .. image:: ../../roboticstoolbox/data/redcar.png
          :width: 300px
          :align: center


        The car is scaled to an image with a length of ``scale`` in the units of
        the plot.

        The vehicle rotates about its ``origin`` which is expressed in terms of
        normalized coordinates in the range 0 to 1.  By default it is in the
        middle of the icon image, (0.2, 0.5) moves it toward the back of the
        vehicle, (0.8, 0.5) moves it toward the front of the vehicle.

        .. note:: 
        
            - The standard icons are kept in ``roboticstoolbox/data``
            - By default the image is assumed to contain a car parallel to
              the x-axis and facing right.  If the vehicle is facing upward
              set ``rotation`` to 90.

        :seealso: :func:`~Vehicle`
        """
        super().__init__()
        if '.' not in filename:
            try:
                # try the default folder first
                image = loaddata(filename + ".png", plt.imread)
            except FileNotFoundError:
                raise ValueError(f"{filename} is not a provided icon")
        else:
            try:
                image = plt.imread(filename)
            except FileNotFoundError:
                raise ValueError(f"icon file {filename} not found")
        
        self._rotation = rotation
        self._image = image

        # figure size of bounding box the image will fill in data coordinates
        if origin is None:
            origin = [0.5, 0.5]
        self._origin = origin

        if image.shape[0] >= image.shape[1]:
            # width >= height
            self._width = scale
            self._height =scale * image.shape[1] / image.shape[0]
        else:
            # width < height
            self._height = scale
            self._width = scale * image.shape[0] / image.shape[1]


    def _add(self, ax=None, **kwargs):

        def imshow_affine(ax, z, *args, **kwargs):
            im = ax.imshow(z, *args, **kwargs, zorder=3)
            x1, x2, y1, y2 = im.get_extent()
            # im._image_skew_coordinate = (x2, y1)
            return im

        self._ax = plt.gca()
        extent = [  -self._origin[0] * self._height, 
                    (1 - self._origin[0]) * self._height, 
                    -self._origin[1] * self._width, 
                    (1 - self._origin[1]) * self._width
                ]
        self._ax = plt.gca()
        self._object = imshow_affine(self._ax, self._image,
                            interpolation='none',
                            extent=extent, clip_on=True,
                            alpha=1)

    def _update(self, x):

        # center_x = self._width // 2
        # center_y = self._height // 2
        center_x = 0
        center_y = 0

        T = (mtransforms.Affine2D()
                    .rotate_deg_around(center_x, center_y, np.degrees(x[2]) - self._rotation)
                    .translate(x[0], x[1])
                    + self._ax.transData)
        self._object.set_transform(T)

if __name__ == "__main__":
    from math import pi

    V = np.diag(np.r_[0.02, 0.5 * pi / 180] ** 2)

    v = VehiclePolygon()
    # v = VehicleIcon('greycar2', scale=2, rotation=90)

    veh = Bicycle(covar=V, animation=v, control=RandomPath(10), verbose=False)
    print(veh)

    odo = veh.step(1, 0.3)
    print(odo)

    print(veh.x)

    print(veh.f([0, 0, 0], odo))

    def control(v, t, x):
        goal = (6,6)
        goal_heading = atan2(goal[1]-x[1], goal[0]-x[0])
        d_heading = base.angdiff(goal_heading, x[2])
        v.stopif(base.norm(x[0:2] - goal) < 0.1)

        return (1, d_heading)

    veh.control=RandomPath(10)
    p = veh.run(1000, plot=True)
    # plt.show()
    print(p)

    veh.plot_xyt_t()
    # veh.plot(p)

    # t, x = veh.path(5, u=control)
    # print(t)

    # fig, ax = plt.subplots()

    # ax.set_xlim(-5, 5)
    # ax.set_ylim(-5, 5)


    # v = VehicleAnimation.Polygon(shape='triangle', maxdim=0.1, color='r')
    # v = VehicleAnimation.Icon('car3.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.Icon('/Users/corkep/Dropbox/code/robotics-toolbox-python/roboticstoolbox/data/car1.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.icon('car3.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.marker()
    # v.start()
    # plt.grid(True)
    # # plt.axis('equal')

    # for theta in np.linspace(0, 2 * np.pi, 100):
    #     v.update([0, 0, theta])
    #     plt.pause(0.1)


# Copyright (c) 2015, Bielefeld University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Bielefeld University
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior
#       written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
import sys

# bold colors
_ansi = {'red': 91, 'yellow': 93}


def is_tty(stream):  # taken from catkin_tools/common.py  # pragma: no cover
    """Returns True if the given stream is a tty, else False"""
    return hasattr(stream, 'isatty') and stream.isatty()


def colorize(msg, color, file=sys.stderr, alt_text=None):  # pragma: no cover
    if color and is_tty(file):
        return '\033[%dm%s\033[0m' % (_ansi[color], msg)
    elif alt_text:
        return '%s%s' % (alt_text, msg)
    else:
        return msg


def message(msg, *args, **kwargs):  # pragma: no cover
    file = kwargs.get('file', sys.stderr)
    alt_text = kwargs.get('alt_text', None)
    color = kwargs.get('color', None)
    print(colorize(msg, color, file, alt_text), *args, file=file)


def warning(*args, **kwargs):  # pragma: no cover
    defaults = dict(file=sys.stderr, alt_text='warning: ', color='yellow')
    defaults.update(kwargs)
    message(*args, **defaults)


def error(*args, **kwargs):  # pragma: no cover
    defaults = dict(file=sys.stderr, alt_text='error: ', color='red')
    defaults.update(kwargs)
    message(*args, **defaults)



import numpy as np
from scipy.linalg import lu


def jsingu(J):

    # convert to row-echelon form
    P, L, U = lu(J)
    U = np.where(abs(U) < 100 * np.finfo(np.float64).eps, False, True)

    for j in range(J.shape[0]):
        if not U[j, j]:
            print(
                f'joint {j} is dependent on joint ' +
                ', '.join(
                    [str(i) for i in range(j) if all(U[:, j] == U[:, i])]))


if __name__ == "__main__":   # pragma nocover
    import roboticstoolbox as rtb

    puma = rtb.models.DH.Puma560()
    J = puma.jacob0(puma.qr)

    jsingu(J)



#!/usr/bin/env python

from math import pi, sin, cos
import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Jaco(DHRobot):
    """
    Class that models a  Kinova Jaco manipulator

    :param symbolic: use symbolic constants
    :type symbolic: bool

    ``Jaco()`` is an object which models a Kinova Jaco robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Puma560()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration
    - qs, arm is stretched out in the x-direction
    - qn, arm is at a nominal non-singular configuration

    .. note::
        - SI units are used.

    :references:
        - "DH Parameters of Jaco" Version 1.0.8, July 25, 2013.

    :seealso: :func:`Mico`
    """

    def __init__(self, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0

        deg = pi / 180    
        # robot length values (metres)
        D1 = 0.2755
        D2 = 0.4100
        D3 = 0.2073
        D4 = 0.0743
        D5 = 0.0743
        D6 = 0.1687
        e2 = 0.0098
        
        # alternate parameters
        aa = 30 * deg
        ca = cos(aa)
        sa = sin(aa)
        c2a = cos(2 * aa)
        s2a = sin(2 * aa)
        d4b = D3 + sa /s2a * D4
        d5b = sa / s2a * D4 + sa / s2a * D5
        d6b = sa / s2a * D5 + D6
        
        # and build a serial link manipulator
        
        # offsets from the table on page 4, "Mico" angles are the passed joint
        # angles.  "DH Algo" are the result after adding the joint angle offset.

        super().__init__(
                    [
                        RevoluteDH(alpha=pi/2,  a=0,  d=D1,   flip=True),
                        RevoluteDH(alpha=pi,    a=D2, d=0,    offset=-pi/2),
                        RevoluteDH(alpha=pi/2,  a=0,  d=-e2,  offset=pi/2),
                        RevoluteDH(alpha=2*aa,  a=0,  d=-d4b),
                        RevoluteDH(alpha=2*aa,  a=0,  d=-d5b, offset=-pi),
                        RevoluteDH(alpha=pi,    a=0,  d=-d6b, offset=100*deg)
                    ],
                    name='Jaco', 
                    manufacturer='Kinova',
                    keywords=('symbolic',)
                )

        self.addconfiguration('qz', np.r_[0, 0, 0, 0, 0, 0]) # zero angles
        self.addconfiguration('qr', np.r_[270, 180, 180, 0, 0, 0]*deg) # vertical pose as per Fig 2

if __name__ == '__main__':    # pragma nocover

    jaco = Jaco(symbolic=False)
    print(jaco)



#!/usr/bin/env python
"""
@author: Peter Corke
"""


# from math import pi
import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Hyper(DHRobot):
    """
    Create model of a hyper redundant planar manipulator

    :param N: number of links, defaults to 10
    :type N: int, optional
    :param a: link length, defaults total ``1/N`` giving a reach of 1
    :type a: int or symbolic, optional
    :param symbolic: [description], defaults to False
    :type symbolic: bool, optional

    - ``Hyper()`` is an object which describes the kinematics of a serial link
      manipulator with 10 joints which moves in the xy-plane, using standard DH
      conventions. At zero angles it forms a straight line along the x-axis. 

    - ``Hyper(N)`` as above, but models a robot with ``N`` joints.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Hyper()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration

    :References: 
    
      - "A divide and conquer articulated-body algorithm for parallel O(log(n))
      calculation of rigid body dynamics, Part 2",
      Int. J. Robotics Research, 18(9), pp 876-892. 

    :seealso: :func:`Coil`, :func:`Ball`

    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self, N=10, a=None, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0

        if a is None:
            a = 1 / N

        links = []
        for i in range(N):
            links.append(RevoluteDH(a=a, alpha=zero))

        super().__init__(
            links,
            name="Hyper" + str(N),
            keywords=('symbolic',),
            symbolic=symbolic
        )

        # zero angles, straight
        self.addconfiguration("qz", np.zeros((N,)))

if __name__ == '__main__':    # pragma nocover

    hyper = Hyper(N=10, symbolic=False)
    print(hyper)

    #print(hyper.fkine(hyper.qz))



import sys
import copy
import numpy as np
import roboticstoolbox as rtb
from spatialmath import SE3
from spatialmath.base.argcheck import isvector, getvector, getmatrix, \
    verifymatrix, getunit
from roboticstoolbox.robot.Link import Link
from spatialmath.base.transforms3d import tr2delta
# from roboticstoolbox.tools import urdf
# from roboticstoolbox.tools import xacro
from pathlib import PurePath, PurePosixPath
from scipy.optimize import minimize, Bounds, LinearConstraint
from roboticstoolbox.tools.null import null
from ansitable import ANSITable, Column

from roboticstoolbox.backends.PyPlot import PyPlot, PyPlot2
from roboticstoolbox.backends.PyPlot.EllipsePlot import EllipsePlot
from roboticstoolbox.backends.Swift import Swift

from roboticstoolbox.robot.Dynamics import DynamicsMixin
from roboticstoolbox.robot.IK import IKMixin

try:
    from matplotlib import colors
    from matplotlib import cm
    _mpl = True
except ImportError:    # pragma nocover
    pass

try:
    import PIL
    _pil_exists = True
except ImportError:    # pragma nocover
    _pil_exists = False

# TODO maybe this needs to be abstract
# ikine functions need: fkine, jacobe, qlim methods from subclass


class Robot(DynamicsMixin, IKMixin):

    _color = True

    def __init__(
            self,
            links,
            name='noname',
            manufacturer='',
            comment='',
            base=None,
            tool=None,
            gravity=None,
            meshdir=None,
            keywords=(),
            symbolic=False):

        self.name = name
        self.manufacturer = manufacturer
        self.comment = comment
        self.symbolic = symbolic
        self.base = base
        self.tool = tool
        self.basemesh = None
        self._reach = None

        if keywords is not None and not isinstance(keywords, (tuple, list)):
            raise TypeError('keywords must be a list or tuple')
        else:
            self.keywords = keywords

        if gravity is None:
            gravity = np.array([0, 0, 9.81])
        self.gravity = gravity

        # validate the links, must be a list of Link subclass objects
        if not isinstance(links, list):
            raise TypeError('The links must be stored in a list.')

        for link in links:
            if not isinstance(link, Link):
                raise TypeError('links should all be Link subclass')
            link._robot = self
        self._links = links

        # Current joint angles of the robot
        self.q = np.zeros(self.n)
        self.qd = np.zeros(self.n)
        self.qdd = np.zeros(self.n)

        self.control_type = 'v'

        self._configdict = {}

        self._dynchanged = True

        # this probably should go down to DHRobot
        if meshdir is not None:
            classpath = sys.modules[self.__module__].__file__
            self.meshdir = PurePath(classpath).parent / PurePosixPath(meshdir)
            self.basemesh = self.meshdir / "link0.stl"
            for j, link in enumerate(self._links, start=1):
                link.mesh = self.meshdir / "link{:d}.stl".format(j)

        # URDF Parser Attempt
        # # Search mesh dir for meshes
        # if urdfdir is not None:
        #     # Parse the URDF to obtain file paths and scales
        #     data = self._get_stl_file_paths_and_scales(urdfdir)
        #     # Obtain the base mesh
        #     self.basemesh = [data[0][0], data[1][0], data[2][0]]
        #     # Save the respective meshes to each link
        #     for idx in range(1, self.n+1):
        #         self._links[idx-1].mesh = [data[0][idx], data[1][idx],
        #         data[2][idx]]
        # else:
        #     self.basemesh = None

    def copy(self):
        return copy.deepcopy(self)

    def __getitem__(self, i):
        """
        Get link (Robot superclass)

        :param i: link number
        :type i: int
        :return: i'th link of robot
        :rtype: Link subclass

        This also supports iterating over each link in the robot object,
        from the base to the tool.

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> print(robot[1]) # print the 2nd link
            >>> print([link.a for link in robot])  # print all the a_j values

        """
        return self._links[i]

    # URDF Parser Attempt
    # @staticmethod
    # def _get_stl_file_paths_and_scales(urdf_path):
    #     data = [[], [], []]  # [ [filenames] , [scales] , [origins] ]
    #
    #     name, ext = splitext(urdf_path)
    #
    #     if ext == '.xacro':
    #         urdf_string = xacro.main(urdf_path)
    #         urdf = URDF.loadstr(urdf_string, urdf_path)
    #
    #         for link in urdf.links:
    #             data[0].append(link.visuals[0].geometry.mesh.filename)
    #             data[1].append(link.visuals[0].geometry.mesh.scale)
    #             data[2].append(SE3(link.visuals[0].origin))
    #
    #     return data

    def dynchanged(self):
        """
        Dynamic parameters have changed (Robot superclass)

        Called from a property setter to inform the robot that the cache of
        dynamic parameters is invalid.

        :seealso: :func:`roboticstoolbox.Link._listen_dyn`
        """
        self._dynchanged = True

    def _getq(self, q=None):
        """
        Get joint coordinates (Robot superclass)

        :param q: passed value, defaults to None
        :type q: array_like, optional
        :return: passed or value from robot state
        :rtype: ndarray(n,)
        """
        if q is None:
            return self.q
        elif isvector(q, self.n):
            return getvector(q, self.n)
        else:
            return getmatrix(q, (None, self.n))

    @property
    def n(self):
        """
        Number of joints (Robot superclass)

        :return: Number of joints
        :rtype: int

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.n

        """
        return self._n

    def addconfiguration(self, name, q, unit='rad'):
        """
        Add a named joint configuration (Robot superclass)

        :param name: Name of the joint configuration
        :type name: str
        :param q: Joint configuration
        :type q: ndarray(n) or list

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.qz
            >>> robot.addconfiguration("mypos", [0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
            >>> robot.mypos
        """
        v = getvector(q, self.n)
        v = getunit(v, unit)
        self._configdict[name] = v
        setattr(self, name, v)

    def configurations_str(self):
        deg = 180 / np.pi

        # TODO: factor this out of DHRobot
        def angle(theta, fmt=None):

            if fmt is not None:
                return fmt.format(theta * deg) + "\u00b0"
            else:  # pragma nocover
                return str(theta * deg) + "\u00b0"

        config = self.config()
        # show named configurations
        if len(self._configdict) > 0:
            table = ANSITable(
                Column("name", colalign=">"),
                *[
                    Column(f"q{j:d}", colalign="<", headalign="<")
                    for j in range(self.n)],
                border="thin")

            for name, q in self._configdict.items():
                qlist = []
                for i, c in enumerate(config):
                    if c == 'P':
                        qlist.append(f"{q[i]: .3g}")
                    else:
                        qlist.append(angle(q[i], "{: .3g}"))
                table.row(name, *qlist)

            return "\n" + str(table)
        else:  # pragma nocover
            return ""

    def linkcolormap(self, linkcolors="viridis"):
        """
        Create a colormap for robot joints

        :param linkcolors: list of colors or colormap, defaults to "viridis"
        :type linkcolors: list or str, optional
        :return: color map
        :rtype: matplotlib.colors.ListedColormap

        - ``cm = robot.linkcolormap()`` is an n-element colormap that gives a
          unique color for every link.  The RGBA colors for link ``j`` are
          ``cm(j)``.
        - ``cm = robot.linkcolormap(cmap)`` as above but ``cmap`` is the name
          of a valid matplotlib colormap.  The default, example above, is the
          ``viridis`` colormap.
        - ``cm = robot.linkcolormap(list of colors)`` as above but a
          colormap is created from a list of n color names given as strings,
          tuples or hexstrings.

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> cm = robot.linkcolormap("inferno")
            >>> print(cm(range(6))) # cm(i) is 3rd color in colormap
            >>> cm = robot.linkcolormap(
            >>>     ['red', 'g', (0,0.5,0), '#0f8040', 'yellow', 'cyan'])
            >>> print(cm(range(6)))

        .. note::

            - Colormaps have 4-elements: red, green, blue, alpha (RGBA)
            - Names of supported colors and colormaps are defined in the
              matplotlib documentation.

                - `Specifying colors
                <https://matplotlib.org/3.1.0/tutorials/colors/colors.html#sphx-glr-tutorials-colors-colors-py>`_
                - `Colormaps
                <https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html#sphx-glr-tutorials-colors-colormaps-py>`_
        """  # noqa

        if isinstance(linkcolors, list) and len(linkcolors) == self.n:
            # provided a list of color names
            return colors.ListedColormap(linkcolors)
        else:
            # assume it is a colormap name
            return cm.get_cmap(linkcolors, 6)

    def manipulability(
            self, q=None, J=None, method='yoshikawa',
            axes='all', **kwargs):
        """
        Manipulability measure

        :param q: Joint coordinates, one of J or q required
        :type q: ndarray(n), or ndarray(m,n)
        :param J: Jacobian in world frame if already computed, one of J or
            q required
        :type J: ndarray(6,n)
        :param method: method to use, "yoshikawa" (default), "condition",
            "minsingular"  or "asada"
        :type method: str
        :param axes: Task space axes to consider: "all" [default],
            "trans" or "rot"
        :type axes: str
        :param kwargs: extra arguments to pass to ``jacob0``
        :return: manipulability
        :rtype: float or ndarray(m)

        - ``manipulability(q)`` is the scalar manipulability index
          for the robot at the joint configuration ``q``.  It indicates
          dexterity, that is, how well conditioned the robot is for motion
          with respect to the 6 degrees of Cartesian motion.  The values is
          zero if the robot is at a singularity.

        Various measures are supported:

        +-------------------+-------------------------------------------------+
        | Measure           |       Description                               |
        +-------------------+-------------------------------------------------+
        | ``"yoshikawa"``   | Volume of the velocity ellipsoid, *distance*    |
        |                   | from singularity [Yoshikawa85]_                 |
        +-------------------+-------------------------------------------------+
        | ``"invcondition"``| Inverse condition number of Jacobian, isotropy  |
        |                   | of the velocity ellipsoid [Klein87]_            |
        +-------------------+-------------------------------------------------+
        | ``"minsingular"`` | Minimum singular value of the Jacobian,         |
        |                   | *distance*  from singularity [Klein87]_         |
        +-------------------+-------------------------------------------------+
        | ``"asada"``       | Isotropy of the task-space acceleration         |
        |                   | ellipsoid which is a function of the Cartesian  |
        |                   | inertia matrix which depends on the inertial    |
        |                   | parameters [Asada83]_                           |
        +-------------------+-------------------------------------------------+

        **Trajectory operation**:

        If ``q`` is a matrix (m,n) then the result (m,) is a vector of
        manipulability indices for each joint configuration specified by a row
        of ``q``.

        .. note::

            - Invokes the ``jacob0`` method of the robot if ``J`` is not passed
            - The "all" option includes rotational and translational
              dexterity, but this involves adding different units. It can be
              more useful to look at the translational and rotational
              manipulability separately.
            - Examples in the RVC book (1st edition) can be replicated by
              using the "all" option
            - Asada's measure requires inertial a robot model with inertial
              parameters.

        :references:

        .. [Yoshikawa85] Manipulability of Robotic Mechanisms. Yoshikawa T.,
                The International Journal of Robotics Research.
                1985;4(2):3-9. doi:10.1177/027836498500400201
        .. [Asada83] A geometrical representation of manipulator dynamics and
                its application to arm design, H. Asada,
                Journal of Dynamic Systems, Measurement, and Control,
                vol. 105, p. 131, 1983.
        .. [Klein87] Dexterity Measures for the Design and Control of
                Kinematically Redundant Manipulators. Klein CA, Blaho BE.
                The International Journal of Robotics Research.
                1987;6(2):72-83. doi:10.1177/027836498700600206

        - Robotics, Vision & Control, Chap 8, P. Corke, Springer 2011.

        """
        if axes == 'all':
            axes = [True, True, True, True, True, True]
        elif axes.startswith('trans'):
            axes = [True, True, True, False, False, False]
        elif axes.startswith('rot'):
            axes = [False, False, False, True, True, True]
        else:
            raise ValueError('axes must be all, trans or rot')

        def yoshikawa(robot, J, q, axes, **kwargs):
            J = J[axes, :]
            if J.shape[0] == J.shape[1]:
                # simplified case for square matrix
                return abs(np.linalg.det(J))
            else:
                m2 = np.linalg.det(J @ J.T)
                return np.sqrt(abs(m2))

        def invcondition(robot, J, q, axes, **kwargs):
            J = J[axes, :]
            return 1 / np.linalg.cond(J)  # return 1/cond(J)

        def minsingular(robot, J, q, axes, **kwargs):
            J = J[axes, :]
            s = np.linalg.svd(J, compute_uv=False)
            return s[-1]  # return last/smallest singular value of J

        def asada(robot, J, q, axes, **kwargs):
            dof = np.sum(axes)
            if np.linalg.matrix_rank(J) < 6:
                return 0
            Ji = np.linalg.pinv(J)
            Mx = Ji.T @ robot.inertia(q) @ Ji
            d = np.where(axes)[0]
            Mx = Mx[d]
            Mx = Mx[:, d.tolist()]
            e, _ = np.linalg.eig(Mx)
            return np.min(e) / np.max(e)

        # choose the handler function
        if method == 'yoshikawa':
            mfunc = yoshikawa
        elif method == 'invcondition':
            mfunc = condition
        elif method == 'minsingular':
            mfunc = minsingular
        elif method == 'asada':
            mfunc = asada
        else:
            raise ValueError(
                "Invalid method chosen")

        # Calculate manipulability based on supplied Jacobian
        if J is not None:
            w = [mfunc(self, J, q, axes)]

        # Otherwise use the q vector/matrix
        else:
            q = getmatrix(q, (None, self.n))
            w = np.zeros(q.shape[0])

            for k, qk in enumerate(q):
                Jk = self.jacob0(qk, **kwargs)
                w[k] = mfunc(self, Jk, qk, axes)

        if len(w) == 1:
            return w[0]
        else:
            return w

# --------------------------------------------------------------------- #

    @property
    def name(self):
        """
        Get/set robot name (Robot superclass)

        - ``robot.name`` is the robot name

        :return: robot name
        :rtype: str

        - ``robot.name = ...`` checks and sets therobot name
        """
        return self._name

    @name.setter
    def name(self, name_new):
        self._name = name_new

# --------------------------------------------------------------------- #

    @property
    def manufacturer(self):
        """
        Get/set robot manufacturer's name (Robot superclass)

        - ``robot.manufacturer`` is the robot manufacturer's name

        :return: robot manufacturer's name
        :rtype: str

        - ``robot.manufacturer = ...`` checks and sets the manufacturer's name
        """
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, manufacturer_new):
        self._manufacturer = manufacturer_new
# --------------------------------------------------------------------- #

    @property
    def links(self):
        """
        Robot links (Robot superclass)

        :return: A list of link objects
        :rtype: list of Link subclass instances

        .. note:: It is probably more concise to index the robot object rather
            than the list of links, ie. the following are equivalent::

                robot.links[i]
                robot[i]
        """
        return self._links

# --------------------------------------------------------------------- #

    @property
    def base(self):
        """
        Get/set robot base transform (Robot superclass)

        - ``robot.base`` is the robot base transform

        :return: robot tool transform
        :rtype: SE3 instance

        - ``robot.base = ...`` checks and sets the robot base transform

        .. note:: The private attribute ``_base`` will be None in the case of
            no base transform, but this property will return ``SE3()`` which
            is an identity matrix.
        """
        if self._base is None:
            return SE3()
        else:
            return self._base

    @base.setter
    def base(self, T):
        # if not isinstance(T, SE3):
        #     T = SE3(T)
        if T is None or isinstance(T, SE3):
            self._base = T
        elif SE3.isvalid(T):
            self._tool = SE3(T, check=False)
        else:
            raise ValueError('base must be set to None (no tool) or an SE3')
# --------------------------------------------------------------------- #

    @property
    def tool(self):
        """
        Get/set robot tool transform (Robot superclass)

        - ``robot.tool`` is the robot name

        :return: robot tool transform
        :rtype: SE3 instance

        - ``robot.tool = ...`` checks and sets the robot tool transform

        .. note:: The private attribute ``_tool`` will be None in the case of
            no tool transform, but this property will return ``SE3()`` which
            is an identity matrix.
        """
        if self._tool is None:
            return SE3()
        else:
            return self._tool

    @tool.setter
    def tool(self, T):
        # if not isinstance(T, SE3):
        #     T = SE3(T)
        # this is allowed to be none, it's helpful for symbolics rather than
        # having an identity matrix
        if T is None or isinstance(T, SE3):
            self._tool = T
        elif SE3.isvalid(T):
            self._tool = SE3(T, check=False)
        else:
            raise ValueError('tool must be set to None (no tool) or an SE3')

    @property
    def qlim(self):
        r"""
        Joint limits (Robot superclass)

        :return: Array of joint limit values
        :rtype: ndarray(2,n)

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.qlim
        """
        # TODO tidy up
        limits = np.zeros((2, self.n))
        for j, link in enumerate(self):
            if link.qlim is None:
                if link.isrevolute:
                    v = np.r_[-np.pi, np.pi]
                else:
                    raise ValueError('undefined prismatic joint limit')
            else:
                v = link.qlim
            limits[:, j] = v
        return limits

# TODO, the remaining functions, I have only a hazy understanding
# of how they work
# --------------------------------------------------------------------- #

    @property
    def q(self):
        """
        Get/set robot joint configuration (Robot superclass)

        - ``robot.q`` is the robot joint configuration

        :return: robot joint configuration
        :rtype: ndarray(n,)

        - ``robot.q = ...`` checks and sets the joint configuration

        .. note::  ???
        """
        return self._q

    @q.setter
    def q(self, q_new):
        self._q = getvector(q_new, self.n)
# --------------------------------------------------------------------- #

    @property
    def qd(self):
        """
        Get/set robot joint velocity (Robot superclass)

        - ``robot.qd`` is the robot joint velocity

        :return: robot joint velocity
        :rtype: ndarray(n,)

        - ``robot.qd = ...`` checks and sets the joint velocity

        .. note::  ???
        """
        return self._qd

    @qd.setter
    def qd(self, qd_new):
        self._qd = getvector(qd_new, self.n)
# --------------------------------------------------------------------- #

    @property
    def qdd(self):
        """
        Get/set robot joint acceleration (Robot superclass)

        - ``robot.qdd`` is the robot joint acceleration

        :return: robot joint acceleration
        :rtype: ndarray(n,)

        - ``robot.qdd = ...`` checks and sets the robot joint acceleration

        .. note::  ???
        """
        return self._qdd

    @qdd.setter
    def qdd(self, qdd_new):
        self._qdd = getvector(qdd_new, self.n)
# --------------------------------------------------------------------- #

# TODO could we change this to control_mode ?
    @property
    def control_type(self):
        """
        Get/set robot control mode (Robot superclass)

        - ``robot.control_type`` is the robot control mode

        :return: robot control mode
        :rtype: ndarray(n,)

        - ``robot.control_type = ...`` checks and sets the robot control mode

        .. note::  ???
        """
        return self._control_type

    @control_type.setter
    def control_type(self, cn):
        if cn == 'p' or cn == 'v' or cn == 'a':
            self._control_type = cn
        else:
            raise ValueError(
                'Control type must be one of \'p\', \'v\', or \'a\'')

# --------------------------------------------------------------------- #

    def plot(
            self, q, backend=None, block=True, dt=0.050,
            limits=None, vellipse=False, fellipse=False,
            jointaxes=True, eeframe=True, shadow=True, name=True, movie=None
            ):
        """
        Graphical display and animation

        :param q: The joint configuration of the robot.
        :type q: float ndarray(n)
        :param backend: The graphical backend to use, currently 'swift'
            and 'pyplot' are implemented. Defaults to 'swift' of an ``ERobot``
            and 'pyplot` for a ``DHRobot``
        :type backend: string
        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param dt: if q is a trajectory, this describes the delay in
            seconds between frames
        :type dt: float
        :param limits: Custom view limits for the plot. If not supplied will
            autoscale, [x1, x2, y1, y2, z1, z2]
            (this option is for 'pyplot' only)
        :type limits: ndarray(6)
        :param vellipse: (Plot Option) Plot the velocity ellipse at the
            end-effector (this option is for 'pyplot' only)
        :type vellipse: bool
        :param vellipse: (Plot Option) Plot the force ellipse at the
            end-effector (this option is for 'pyplot' only)
        :type vellipse: bool
        :param jointaxes: (Plot Option) Plot an arrow indicating the axes in
            which the joint revolves around(revolute joint) or translates
            along (prosmatic joint) (this option is for 'pyplot' only)
        :type jointaxes: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
            (this option is for 'pyplot' only)
        :type eeframe: bool
        :param shadow: (Plot Option) Plot a shadow of the robot in the x-y
            plane. (this option is for 'pyplot' only)
        :type shadow: bool
        :param name: (Plot Option) Plot the name of the robot near its base
            (this option is for 'pyplot' only)
        :type name: bool
        :param movie: name of file in which to save an animated GIF
            (this option is for 'pyplot' only)
        :type movie: str

        :return: A reference to the environment object which controls the
            figure
        :rtype: Swift or PyPlot

        - ``robot.plot(q, 'pyplot')`` displays a graphical view of a robot
          based on the kinematic model and the joint configuration ``q``.
          This is a stick figure polyline which joins the origins of the
          link coordinate frames. The plot will autoscale with an aspect
          ratio of 1.

        If ``q`` (m,n) representing a joint-space trajectory it will create an
        animation with a pause of ``dt`` seconds between each frame.

        .. note::
            - By default this method will block until the figure is dismissed.
              To avoid this set ``block=False``.
            - For PyPlot, the polyline joins the origins of the link frames,
              but for some Denavit-Hartenberg models those frames may not
              actually be on the robot, ie. the lines to not neccessarily
              represent the links of the robot.

        :seealso: :func:`teach`
        """

        env = None

        if backend is None:
            if isinstance(self, rtb.DHRobot):
                backend = 'pyplot'
            else:
                backend = 'swift'

        if backend.lower() == 'swift':  # pragma nocover
            if isinstance(self, rtb.ERobot):
                env = self._plot_swift(q=q, block=block)
            elif isinstance(self, rtb.DHRobot):
                raise NotImplementedError(
                    'Plotting in Swift is not implemented for DHRobots yet')

        elif backend.lower() == 'pyplot':
            # if isinstance(self, rtb.ERobot):  # pragma nocover
            #     raise NotImplementedError(
            #         'Plotting in PyPlot is not implemented for ERobots yet')
            # elif isinstance(self, rtb.DHRobot):
            env = self._plot_pyplot(
                q=q, block=block, dt=dt, limits=limits, vellipse=vellipse,
                fellipse=fellipse, jointaxes=jointaxes, eeframe=eeframe,
                shadow=shadow, name=name, movie=movie)

        return env

    def _plot_pyplot(
            self, q, block, dt, limits,
            vellipse, fellipse,
            jointaxes, eeframe, shadow, name, movie):

        # Make an empty 3D figure
        env = PyPlot()

        q = getmatrix(q, (None, self.n))

        # Add the self to the figure in readonly mode
        if q.shape[0] == 1:
            env.launch(self.name + ' Plot', limits)
        else:
            env.launch(self.name + ' Trajectory Plot', limits)

        env.add(
            self, readonly=True,
            jointaxes=jointaxes, eeframe=eeframe, shadow=shadow, name=name)

        if vellipse:
            vell = self.vellipse(centre='ee')
            env.add(vell)

        if fellipse:
            fell = self.fellipse(centre='ee')
            env.add(fell)

        # Stop lint error
        images = []  # list of images saved from each plot

        if movie is not None:   # pragma nocover
            if not _pil_exists:
                raise RuntimeError(
                    'to save movies PIL must be installed:\npip3 install PIL')
            # make the background white, looks better than grey stipple
            env.ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
            env.ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
            env.ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))

        for qk in q:
            self.q = qk
            env.step(dt)

            if movie is not None:  # pragma nocover
                # render the frame and save as a PIL image in the list
                canvas = env.fig.canvas
                img = PIL.Image.frombytes(
                    'RGB', canvas.get_width_height(),
                    canvas.tostring_rgb())
                images.append(img)

        if movie is not None:  # pragma nocover
            # save it as an animated GIF
            images[0].save(
                movie,
                save_all=True, append_images=images[1:], optimize=False,
                duration=dt, loop=0)

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

    def _plot_swift(self, q, block):   # pragma nocover

        # Make an empty 3D figure
        env = Swift()

        q = getmatrix(q, (None, self.n))
        self.q = q[0, :]

        # Add the self to the figure in readonly mode
        env.launch()

        env.add(
            self, readonly=True)

        for qk in q:
            self.q = qk
            env.step()

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

# --------------------------------------------------------------------- #

    def fellipse(self, q=None, opt='trans', centre=[0, 0, 0]):
        '''
        Create a force ellipsoid object for plotting with PyPlot

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param opt: 'trans' or 'rot' will plot either the translational or
            rotational force ellipsoid
        :type opt: string
        :param centre:
        :type centre: list or str('ee')

        :return: An EllipsePlot object
        :rtype: EllipsePlot

        - ``robot.fellipse(q)`` creates a force ellipsoid for the robot at
          pose ``q``. The ellipsoid is centered at the origin.

        - ``robot.fellipse()`` as above except the joint configuration is that
          stored in the robot object.

        .. note::
            - By default the ellipsoid related to translational motion is
              drawn.  Use ``opt='rot'`` to draw the rotational velocity
              ellipsoid.
            - By default the ellipsoid is drawn at the origin.  The option
              ``centre`` allows its origin to set to set to the specified
              3-vector, or the string "ee" ensures it is drawn at the
              end-effector position.

        '''
        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "ERobot fellipse not implemented yet")

        ell = EllipsePlot(self, q, 'f', opt, centre=centre)
        return ell

    def vellipse(self, q=None, opt='trans', centre=[0, 0, 0]):
        """
        Create a velocity ellipsoid object for plotting with PyPlot

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param opt: 'trans' or 'rot' will plot either the translational or
            rotational velocity ellipsoid
        :type opt: string
        :param centre:
        :type centre: list or str('ee')

        :return: An EllipsePlot object
        :rtype: EllipsePlot

        - ``robot.vellipse(q)`` creates a force ellipsoid for the robot at
          pose ``q``. The ellipsoid is centered at the origin.

        - ``robot.vellipse()`` as above except the joint configuration is that
          stored in the robot object.

        .. note::
            - By default the ellipsoid related to translational motion is
              drawn.  Use ``opt='rot'`` to draw the rotational velocity
              ellipsoid.
            - By default the ellipsoid is drawn at the origin.  The option
              ``centre`` allows its origin to set to set to the specified
              3-vector, or the string "ee" ensures it is drawn at the
              end-effector position.
        """
        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "ERobot vellipse not implemented yet")

        ell = EllipsePlot(self, q, 'v', opt, centre=centre)
        return ell

    def plot_ellipse(
            self, ellipse, block=True, limits=None,
            jointaxes=True, eeframe=True, shadow=True, name=True):
        """
        Plot the an ellipsoid

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param ellipse: the ellipsoid to plot
        :type ellipse: EllipsePlot
        :param jointaxes: (Plot Option) Plot an arrow indicating the axes in
            which the joint revolves around(revolute joint) or translates
            along (prosmatic joint)
        :type jointaxes: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param shadow: (Plot Option) Plot a shadow of the robot in the x-y
            plane
        :type shadow: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.plot_ellipse(ellipsoid)`` displays the ellipsoid.

        .. note::
            - By default the ellipsoid is drawn at the origin.  The option
              ``centre`` allows its origin to set to set to the specified
              3-vector, or the string "ee" ensures it is drawn at the
              end-effector position.
        """

        if not isinstance(ellipse, EllipsePlot):  # pragma nocover
            raise TypeError(
                'ellipse must be of type '
                'roboticstoolbox.backend.PyPlot.EllipsePlot')

        env = PyPlot()

        # Add the robot to the figure in readonly mode
        env.launch(ellipse.robot.name + ' ' + ellipse.name, limits=limits)

        env.add(
            ellipse,
            jointaxes=jointaxes, eeframe=eeframe, shadow=shadow, name=name)

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

    def plot_fellipse(
            self, q=None, block=True, fellipse=None,
            limits=None, opt='trans', centre=[0, 0, 0],
            jointaxes=True, eeframe=True, shadow=True, name=True):
        """
        Plot the force ellipsoid for manipulator

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param fellipse: the vellocity ellipsoid to plot
        :type fellipse: EllipsePlot
        :param limits: Custom view limits for the plot. If not supplied will
            autoscale, [x1, x2, y1, y2, z1, z2]
        :type limits: ndarray(6)
        :param opt: 'trans' or 'rot' will plot either the translational or
            rotational force ellipsoid
        :type opt: string
        :param centre: The coordinates to plot the fellipse [x, y, z] or "ee"
            to plot at the end-effector location
        :type centre: array_like or str
        :param jointaxes: (Plot Option) Plot an arrow indicating the axes in
            which the joint revolves around(revolute joint) or translates
            along (prosmatic joint)
        :type jointaxes: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param shadow: (Plot Option) Plot a shadow of the robot in the x-y
            plane
        :type shadow: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.plot_fellipse(q)`` displays the velocity ellipsoid for the
          robot at pose ``q``. The plot will autoscale with an aspect ratio
          of 1.

        - ``plot_fellipse()`` as above except the robot is plotted with joint
          coordinates stored in the robot object.

        - ``robot.plot_fellipse(vellipse)`` specifies a custon ellipse to plot.

        .. note::
            - By default the ellipsoid related to translational motion is
              drawn.  Use ``opt='rot'`` to draw the rotational velocity
              ellipsoid.
            - By default the ellipsoid is drawn at the origin.  The option
              ``centre`` allows its origin to set to set to the specified
              3-vector, or the string "ee" ensures it is drawn at the
              end-effector position.
        """

        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "Ellipse Plotting of ERobot's not implemented yet")

        if q is not None:
            self.q = q

        if fellipse is None:
            fellipse = self.fellipse(q=q, opt=opt, centre=centre)

        return self.plot_ellipse(
            fellipse, block=block, limits=limits,
            jointaxes=jointaxes, eeframe=eeframe, shadow=shadow, name=name)

    def plot_vellipse(
            self, q=None, block=True, vellipse=None,
            limits=None, opt='trans', centre=[0, 0, 0],
            jointaxes=True, eeframe=True, shadow=True, name=True):
        """
        Plot the velocity ellipsoid for manipulator

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param vellipse: the vellocity ellipsoid to plot
        :type vellipse: EllipsePlot
        :param limits: Custom view limits for the plot. If not supplied will
            autoscale, [x1, x2, y1, y2, z1, z2]
        :type limits: ndarray(6)
        :param opt: 'trans' or 'rot' will plot either the translational or
            rotational velocity ellipsoid
        :type opt: string
        :param centre: The coordinates to plot the vellipse [x, y, z] or "ee"
            to plot at the end-effector location
        :type centre: array_like or str
        :param jointaxes: (Plot Option) Plot an arrow indicating the axes in
            which the joint revolves around(revolute joint) or translates
            along (prosmatic joint)
        :type jointaxes: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param shadow: (Plot Option) Plot a shadow of the robot in the x-y
            plane
        :type shadow: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.plot_vellipse(q)`` displays the velocity ellipsoid for the
          robot at pose ``q``. The plot will autoscale with an aspect ratio
          of 1.

        - ``plot_vellipse()`` as above except the robot is plotted with joint
          coordinates stored in the robot object.

        - ``robot.plot_vellipse(vellipse)`` specifies a custon ellipse to plot.

        .. note::
            - By default the ellipsoid related to translational motion is
              drawn.  Use ``opt='rot'`` to draw the rotational velocity
              ellipsoid.
            - By default the ellipsoid is drawn at the origin.  The option
              ``centre`` allows its origin to set to set to the specified
              3-vector, or the string "ee" ensures it is drawn at the
              end-effector position.
        """

        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "Ellipse Plotting of ERobot's not implemented yet")

        if q is not None:
            self.q = q

        if vellipse is None:
            vellipse = self.vellipse(q=q, opt=opt, centre=centre)

        return self.plot_ellipse(
            vellipse, block=block, limits=limits,
            jointaxes=jointaxes, eeframe=eeframe, shadow=shadow, name=name)

# --------------------------------------------------------------------- #

    def plot2(
            self, q, block=True, dt=0.05, limits=None,
            vellipse=False, fellipse=False,
            eeframe=True, name=False):
        """
        2D Graphical display and animation

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param dt: if q is a trajectory, this describes the delay in
            milliseconds between frames
        :type dt: int
        :param limits: Custom view limits for the plot. If not supplied will
            autoscale, [x1, x2, y1, y2, z1, z2]
        :type limits: ndarray(6)
        :param vellipse: (Plot Option) Plot the velocity ellipse at the
            end-effector
        :type vellipse: bool
        :param vellipse: (Plot Option) Plot the force ellipse at the
            end-effector
        :type vellipse: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.plot2(q)`` displays a 2D graphical view of a robot based on
          the kinematic model and the joint configuration ``q``. This is a
          stick figure polyline which joins the origins of the link coordinate
          frames. The plot will autoscale with an aspect ratio of 1.

        If ``q`` (m,n) representing a joint-space trajectory it will create an
        animation with a pause of ``dt`` seconds between each frame.

        .. note::
            - By default this method will block until the figure is dismissed.
              To avoid this set ``block=False``.
            - The polyline joins the origins of the link frames, but for
              some Denavit-Hartenberg models those frames may not actually
              be on the robot, ie. the lines to not neccessarily represent
              the links of the robot.

        :seealso: :func:`teach2`

        """

        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "2D Plotting of ERobot's not implemented yet")

        # Make an empty 2D figure
        env = PyPlot2()

        q = getmatrix(q, (None, self.n))

        # Add the self to the figure in readonly mode
        if q.shape[0] == 1:
            env.launch(self.name + ' Plot', limits)
        else:
            env.launch(self.name + ' Trajectory Plot', limits)

        env.add(
            self, readonly=True,
            eeframe=eeframe, name=name)

        if vellipse:
            vell = self.vellipse(centre='ee')
            env.add(vell)

        if fellipse:
            fell = self.fellipse(centre='ee')
            env.add(fell)

        for qk in q:
            self.q = qk
            env.step()

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

# --------------------------------------------------------------------- #

    def teach(
            self, q=None, block=True, order='xyz', limits=None,
            jointaxes=True, eeframe=True, shadow=True, name=True):
        """
        Graphical teach pendant

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param q: The joint configuration of the robot (Optional,
                  if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param limits: Custom view limits for the plot. If not supplied will
                       autoscale, [x1, x2, y1, y2, z1, z2]
        :type limits: ndarray(6)
        :param jointaxes: (Plot Option) Plot an arrow indicating the axes in
                          which the joint revolves around(revolute joint) or
                          translates along (prismatic joint)
        :type jointaxes: bool
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param shadow: (Plot Option) Plot a shadow of the robot in the x-y
            plane
        :type shadow: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.teach(q)`` creates a matplotlib plot which allows the user to
          "drive" a graphical robot using a graphical slider panel. The robot's
          inital joint configuration is ``q``. The plot will autoscale with an
          aspect ratio of 1.

        - ``robot.teach()`` as above except the robot's stored value of ``q``
            is used.

        .. note::
            - Program execution is blocked until the teach window is
              dismissed.  If ``block=False`` the method is non-blocking but
              you need to poll the window manager to ensure that the window
              remains responsive.
            - The slider limits are derived from the joint limit properties.
              If not set then:
                - For revolute joints they are assumed to be [-pi, +pi]
                - For prismatic joint they are assumed unknown and an error
                  occurs.
        """

        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "2D Plotting of ERobot's not implemented yet")

        if q is not None:
            self.q = q

        # Make an empty 3D figure
        env = PyPlot()

        # Add the self to the figure in readonly mode
        env.launch('Teach ' + self.name, limits=limits)
        env.add(
            self, readonly=True,
            jointaxes=jointaxes, eeframe=eeframe, shadow=shadow, name=name)

        env._add_teach_panel(self)

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

    def teach2(
            self, q=None, block=True, limits=None,
            eeframe=True, name=False):
        '''
        2D Graphical teach pendant

        :param block: Block operation of the code and keep the figure open
        :type block: bool
        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param limits: Custom view limits for the plot. If not supplied will
            autoscale, [x1, x2, y1, y2, z1, z2]
        :type limits: ndarray(6)
        :param eeframe: (Plot Option) Plot the end-effector coordinate frame
            at the location of the end-effector. Uses three arrows, red,
            green and blue to indicate the x, y, and z-axes.
        :type eeframe: bool
        :param name: (Plot Option) Plot the name of the robot near its base
        :type name: bool

        :return: A reference to the PyPlot object which controls the
            matplotlib figure
        :rtype: PyPlot

        - ``robot.teach2(q)`` creates a 2D matplotlib plot which allows the
          user to "drive" a graphical robot using a graphical slider panel.
          The robot's inital joint configuration is ``q``. The plot will
          autoscale with an aspect ratio of 1.

        - ``robot.teach2()`` as above except the robot's stored value of ``q``
          is used.

        .. note::
            - Program execution is blocked until the teach window is
              dismissed.  If ``block=False`` the method is non-blocking but
              you need to poll the window manager to ensure that the window
              remains responsive.
            - The slider limits are derived from the joint limit properties.
              If not set then:
                - For revolute joints they are assumed to be [-pi, +pi]
                - For prismatic joint they are assumed unknown and an error
                  occurs.
              If not set then
                - For revolute joints they are assumed to be [-pi, +pi]
                - For prismatic joint they are assumed unknown and an error
                  occurs.

        '''

        if isinstance(self, rtb.ERobot):  # pragma nocover
            raise NotImplementedError(
                "2D Plotting of ERobot's not implemented yet")

        if q is not None:
            self.q = q

        # Make an empty 3D figure
        env = PyPlot2()

        # Add the robot to the figure in readonly mode
        env.launch('Teach ' + self.name, limits=limits)
        env.add(
            self, readonly=True,
            eeframe=eeframe, name=name)

        env._add_teach_panel(self)

        # Keep the plot open
        if block:           # pragma: no cover
            env.hold()

        return env

# --------------------------------------------------------------------- #

    def closest_point(self, shape, inf_dist=1.0):
        '''
        closest_point(shape, inf_dist) returns the minimum euclidean
        distance between this robot and shape, provided it is less than
        inf_dist. It will also return the points on self and shape in the
        world frame which connect the line of length distance between the
        shapes. If the distance is negative then the shapes are collided.
        :param shape: The shape to compare distance to
        :type shape: Shape
        :param inf_dist: The minimum distance within which to consider
            the shape
        :type inf_dist: float
        :returns: d, p1, p2 where d is the distance between the shapes,
            p1 and p2 are the points in the world frame on the respective
            shapes
        :rtype: float, SE3, SE3
        '''

        d = 10000
        p1 = None,
        p2 = None

        for link in self.links:
            td, tp1, tp2 = link.closest_point(shape, inf_dist)

            if td is not None and td < d:
                d = td
                p1 = tp1
                p2 = tp2

        if d == 10000:
            d = None

        return d, p1, p2

    def collided(self, shape):
        '''
        collided(shape) checks if this robot and shape have collided
        :param shape: The shape to compare distance to
        :type shape: Shape
        :returns: True if shapes have collided
        :rtype: bool
        '''

        for link in self.links:
            if link.collided(shape):
                return True

        return False

    def joint_velocity_damper(self, ps=0.05, pi=0.1, n=None, gain=1.0):
        '''
        Formulates an inequality contraint which, when optimised for will
        make it impossible for the robot to run into joint limits. Requires
        the joint limits of the robot to be specified. See examples/mmc.py
        for use case

        :param ps: The minimum angle (in radians) in which the joint is
            allowed to approach to its limit
        :type ps: float
        :param pi: The influence angle (in radians) in which the velocity
            damper becomes active
        :type pi: float
        :param n: The number of joints to consider. Defaults to all joints
        :type n: int
        :param gain: The gain for the velocity damper
        :type gain: float

        :returns: Ain, Bin as the inequality contraints for an optisator
        :rtype: ndarray(6), ndarray(6)
        '''

        if n is None:
            n = self.n

        Ain = np.zeros((n, n))
        Bin = np.zeros(n)

        for i in range(n):
            if self.q[i] - self.qlim[0, i] <= pi:
                Bin[i] = -gain * (
                    ((self.qlim[0, i] - self.q[i]) + ps) / (pi - ps))
                Ain[i, i] = -1
            if self.qlim[1, i] - self.q[i] <= pi:
                Bin[i] = gain * (
                    (self.qlim[1, i] - self.q[i]) - ps) / (pi - ps)
                Ain[i, i] = 1

        return Ain, Bin



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ERobot import ERobot


class wx250s(ERobot):
    """
    Class that imports a wx250s URDF model

    ``wx250s()`` is a class which imports an Interbotix wx250s robot definition
    from a URDF file.  The model describes its kinematic and graphical
    characteristics.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.URDF.wx250s()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration

    :reference:
        - http://www.support.interbotix.com/html/specifications/wx2506dof.html

    .. codeauthor:: Jesse Haviland
    .. sectionauthor:: Peter Corke
    """
    def __init__(self):

        args = super().urdf_to_ets_args(
            "interbotix_descriptions/urdf/wx250s.urdf.xacro")

        super().__init__(
                args[0],
                name=args[1],
                manufacturer='Interbotix'
            )

        self.addconfiguration(
            "qz", np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
        self.addconfiguration(
            "qr", np.array([0, -0.3, 0, -2.2, 0, 2.0, np.pi/4, 0, 0]))


if __name__ == '__main__':   # pragma nocover

    robot = wx250s()
    print(robot)



from setuptools import setup, find_packages, Extension
import os

here = os.path.abspath(os.path.dirname(__file__))

req = [
    'numpy>=1.18.0',
    'spatialmath-python>=0.8.7',
    'pgraph-python',
    'scipy',
    'matplotlib',
    'ansitable',
    'swift-sim>=0.6.1',
    'qpsolvers'
]

collision_req = [
    'pybullet'
]

vp_req = [
    'vpython',
    'numpy-stl',
    'imageio',
    'imageio-ffmpeg'
]

dev_req = [
    'pytest',
    'pytest-cov',
    'flake8',
    'pyyaml',
    'sympy'
]

docs_req = [
    'sphinx',
    'sphinx_rtd_theme',
    'sphinx-autorun',
]

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# list all data folders here, to ensure they get packaged

data_folders = [
    'roboticstoolbox/models/URDF/xacro',
    'roboticstoolbox/models/DH/meshes',
    'roboticstoolbox/data',
]


def package_files(directory):
    paths = []
    for (pathhere, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', pathhere, filename))
    return paths


extra_files = []
for data_folder in data_folders:
    extra_files += package_files(data_folder)

frne = Extension(
        'frne',
        sources=[
            './roboticstoolbox/core/vmath.c',
            './roboticstoolbox/core/ne.c',
            './roboticstoolbox/core/frne.c'])

setup(
    name='roboticstoolbox-python',

    version='0.7.0',

    description='A Python library for robotic education and research',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/petercorke/robotics-toolbox-python',

    author='Jesse Haviland and Peter Corke',

    license='MIT',

    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    python_requires='>=3.6',

    project_urls={
        'Documentation': 'https://petercorke.github.io/roboticstoolbox-python',
        'Source': 'https://github.com/petercorke/roboticstoolbox-python',
        'Tracker': 'https://github.com/petercorke/roboticstoolbox-python/issues',
        'Coverage': 'https://codecov.io/gh/petercorke/roboticstoolbox-python'
    },

    ext_modules=[frne],

    keywords='python robotics robotics-toolbox kinematics dynamics' \
             ' motion-planning trajectory-generation jacobian hessian' \
             ' control simulation robot-manipulator mobile-robot',

    packages=find_packages(exclude=["tests", "examples", "notebooks"]),

    package_data={'roboticstoolbox': extra_files},

    include_package_data=True,

    install_requires=req,

    extras_require={
        'collision': collision_req,
        'dev': dev_req,
        'docs': docs_req,
        'vpython': vp_req
    }
)
